"""Collector plugin for docker images."""

from __future__ import annotations

import json
import re
import urllib.parse

from base64 import b64encode
from pathlib import Path
from subprocess import CalledProcessError
from types import NoneType
from typing import TYPE_CHECKING, Final

import jmespath

from oras.container import Container
from pydantic import SecretStr
from requests import HTTPError

import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.constants import BomProps
from hoppr.core_plugins.oras_registry import Registry
from hoppr.exceptions import HopprPluginError
from hoppr.models.sbom import Property
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result

if TYPE_CHECKING:
    from packageurl import PackageURL

    from hoppr.models import HopprContext
    from hoppr.models.credentials import CredentialRequiredService
    from hoppr.models.sbom import Component


ALLOWED_MEDIA_TYPES: Final[str] = (
    "application/vnd.oci.image.index.v1+json,"
    "application/vnd.docker.distribution.manifest.v2+json,"
    "application/vnd.oci.image.manifest.v1+json"
)

_ANNOTATION_BUNDLE = "dev.sigstore.cosign/bundle"
_ANNOTATION_CERTIFICATE = "dev.sigstore.cosign/certificate"
_ANNOTATION_CHAIN = "dev.sigstore.cosign/chain"
_ANNOTATION_SIGNATURE = "dev.cosignproject.cosign/signature"

_ANNOTATION_EXT_MAP: Final[dict] = {
    _ANNOTATION_BUNDLE: ".bundle",
    _ANNOTATION_CHAIN: ".chain",
}

_ANNOTATION_JSON_EXT_MAP: Final[dict] = {
    _ANNOTATION_SIGNATURE: ".sig",
    _ANNOTATION_CERTIFICATE: ".pem",
}


class CollectDockerPlugin(
    SerialCollectorPlugin,
    supported_purl_types=["docker", "oci"],
    required_commands=["skopeo"],
    products=["docker/*", "oci/*"],
    process_timeout=300,
    system_repositories=["https://docker.io/"],
):
    """Collector plugin for docker images."""

    def get_version(self) -> str:  # noqa: D102
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)
        self.required_commands = (self.config or {}).get("skopeo_command", self.required_commands)

    def _download_package(
        self,
        purl: PackageURL,
        repo_url: str,
        package_str: str,
        dest_file: Path,
        creds: CredentialRequiredService | None,
    ):
        """Downloads a Docker package from a specific repo via Skopeo.

        Args:
            purl: Used to identify the package in the logs and check against purl version
            repo_url: Designates registry to pull package from
            package_str: Docker container to be downloaded.
            dest_file: Where the package will be downloaded to.
            creds: Credentials used for the HTTP request.

        Raises:
            HopprPluginError: If package download fails.
        """
        command = [
            self.required_commands[0],
            "copy",
            "--additional-tag",
            f"{purl.name}:{purl.version.replace(':', '')}",
        ]
        password_list = []

        if creds is not None and isinstance(creds.password, SecretStr):
            password_list = [creds.password.get_secret_value()]
            command = [*command, "--src-creds", f"{creds.username}:{creds.password.get_secret_value()}"]

        if re.match("^http://", repo_url):
            command = [*command, "--src-tls-verify=false"]

        if self.get_logger().is_verbose():
            command = [*command, "--debug"]

        command = [
            *command,
            urllib.parse.unquote(package_str),
            f"{purl.type}-archive:{dest_file}",
        ]
        copy_result = self.run_command(command, password_list)

        try:
            copy_result.check_returncode()
        except CalledProcessError as ex:
            msg = f"Skopeo failed to copy {purl.type} image to {dest_file}, return_code={copy_result.returncode}"
            self.get_logger().error(msg=msg, indent_level=2)
            raise HopprPluginError(msg) from ex

    def _prep_headers(self, client: Registry, creds: CredentialRequiredService | None) -> dict:
        client.set_header("Authorization", None)
        auth_header = {"Authorization": None}
        accept_header = {"Accept": ALLOWED_MEDIA_TYPES}

        if creds is not None and isinstance(creds.password, SecretStr):
            client.auth.set_basic_auth(creds.username, creds.password.get_secret_value())
            basic_auth = b64encode(f"{(creds.username)}:{creds.password.get_secret_value()}".encode()).decode("utf-8")
            client.set_header("Authorization", f"Basic {basic_auth}")

            auth_header = client.auth.get_auth_header()

        return accept_header | auth_header

    def _get_digest(self, client: Registry, container: Container, creds: CredentialRequiredService | None) -> str:
        """Retrieves Digest for Docker package from a specific registry.

        Args:
            client: Oras Registry Object used to call do_request
            container: Oras Container Object of Docker Package
            creds: User Configured Credentials for making requests

        Raises:
            HopprPluginError: If digest retrieval fails.
        """
        headers = self._prep_headers(client=client, creds=creds)

        get_digest = (
            f"{client.prefix}://{container.registry}/v2/{container.api_prefix}"
            f"/manifests/{container.digest or container.tag}"
        )

        try:
            response = client.do_request(get_digest, "GET", headers=headers)
            response.raise_for_status()
            return response.headers.get("Docker-Content-Digest") or ""
        except (ValueError, HTTPError) as ex:
            msg = f"Failed to retrieve Digest for {container.repository}"
            self.get_logger().warning(msg=msg, indent_level=2)
            raise HopprPluginError(msg) from ex

    def _get_signature_manifest(
        self, client: Registry, container: Container, digest: str, creds: CredentialRequiredService | None
    ) -> str:
        """Retrieves signature Manifest for Docker package from a specific registry.

        Returns empty string if no signature manifest found.

        Args:
            client: Oras Registry Object used to call do_request
            container: Oras Container Object of Docker Package
            digest: Digest string for Docker package
            creds: User Configured Credentials for making requests
        """
        headers = self._prep_headers(client=client, creds=creds)
        digest = digest.replace(":", "-")
        get_signature_manifest = (
            f"{client.prefix}://{container.registry}/v2/{container.api_prefix}/manifests/{digest}.sig"
        )
        response = client.do_request(get_signature_manifest, "GET", headers=headers)
        return response.content if response.status_code == 200 else ""

    def _get_oras_container(self, purl: PackageURL, repo_url: str) -> Container:
        """Parses Purl/Repo information to create Oras container Object.

        Args:
            purl: PackageURL for Docker Component
            repo_url: Repo for Docker Package
        """
        version = re.sub(
            pattern=r"^(sha256:)?([a-f0-9]{64})$", repl=r"sha256:\2", string=urllib.parse.unquote(purl.version)
        )

        if (repo := re.sub(r"^https?://", "", repo_url)) == "docker.io":
            return Container(name=f"index.{repo}/{purl.namespace or 'library'}/{purl.name}@{version}")

        return Container(name=f"{repo}/{f'{purl.namespace}/' if purl.namespace else ''}{purl.name}@{version}")

    def _parse_manifest(self, manifest: str, dest_file: Path, digest: str):
        """Parses signature manifest file for Docker package to check for signature entries.

        Args:
            manifest: manifest file containing signature blob shas
            dest_file: path to download files to
            digest: digest to write to json file
        """
        manifest_json = json.loads(manifest)
        # Get all signature layers
        layers = jmespath.search(
            expression="layers[? mediaType == 'application/vnd.dev.cosign.simplesigning.v1+json']",
            data=manifest_json,
        )

        for layer in layers:
            # Get the signature layer's index in the manifest's full `layers` list
            idx = manifest_json.get("layers", []).index(layer)
            self._create_signature_files(layer.get("annotations", {}), dest_file, idx, digest)

    def _create_signature_files(self, annotations_field: dict[str, str], dest_file: Path, index: int, digest: str):
        """Creates signature related files from annotation fields if appropriate.

        Args:
            annotations_field: json encoded string of signature manifest layer annotation field
            index: differentiates multiple of the same file types if more than on signature is present
            dest_file: path to download files to
            digest: digest to write to json file
        """
        qualifier = f"-{index}" if index else ""
        file = f"{dest_file}{qualifier}"

        for field, ext in _ANNOTATION_EXT_MAP.items():
            if found_value := annotations_field.get(field):
                Path(f"{file}{ext}").write_text(found_value, encoding="utf-8")

        for field, ext in _ANNOTATION_JSON_EXT_MAP.items():
            if found_value := annotations_field.get(field):
                Path(f"{file}.json{ext}").write_text(found_value, encoding="utf-8")

        Path(f"{file}.json").write_text(digest, encoding="utf-8")

    def _get_package_signatures(
        self,
        repo_url: str,
        purl: PackageURL,
        dest_file: Path,
        creds: CredentialRequiredService | None,
    ):
        """Attempts to collect signature files for a given component.

        Args:
            repo_url: Repo for Docker Package
            purl: PackageURL for Docker Component
            dest_file: Path of file to write signature to
            creds: Credentials to use with registry.

        Raises:
            HopprPluginError: If collection fails, but a lack of signatures is not a failure
        """
        client = Registry(hostname=re.sub(r"^https?://", "", repo_url), auth_backend="basic", logger=self.get_logger())
        container = self._get_oras_container(purl, repo_url)

        try:
            if digest := self._get_digest(client, container, creds):
                if signature_manifest := self._get_signature_manifest(client, container, digest, creds):
                    for layer in jmespath.search(expression="layers", data=json.loads(signature_manifest)):
                        blob = (
                            client.get_blob(
                                container=container,
                                digest=layer.get("digest"),
                            )
                        ).text
                        self._parse_manifest(signature_manifest, dest_file, blob)
                else:
                    self.get_logger().info(f"No signatures to collect for {purl}.")
        except HopprPluginError:
            self.get_logger().warning(f"Failed to collect signatures for {purl}.")

    def _get_image(self, url: str, purl: PackageURL) -> RepositoryUrl:
        """Return the image details for skopeo to process.

        Args:
            url (str): Repository URL
            purl (PackageURL): Purl of component to operate on

        Returns:
            RepositoryUrl: Image information
        """
        # Determine if purl version contains SHA string, determine proper formatting for skopeo command
        match purl.version:
            case NoneType():
                raise HopprPluginError(f"Missing `version` from purl string: {purl}. Must be a valid tag or digest")
            case no_sha if re.search(r"^(sha256:)?[a-fA-F0-9]{12,64}$", no_sha) is None:
                image_name = f"{purl.name}:{purl.version}"
            case sha if sha.startswith("sha256:"):
                image_name = f"{purl.name}@{purl.version}"
            case _:
                image_name = f"{purl.name}@sha256:{purl.version}"

        if purl.type == "oci" and "repository_url" in purl.qualifiers:
            url = purl.qualifiers.get("repository_url", "")

        source_image = RepositoryUrl(url=url) / (purl.namespace or "") / urllib.parse.quote_plus(image_name)

        if source_image.scheme != "docker":
            source_image = RepositoryUrl(url="docker://" + re.sub(r"^(.*://)", "", str(source_image)))

        return source_image

    def _get_target_path(self, repo_url: str, purl: PackageURL) -> Path:
        """Get target path for image download.

        Args:
            repo_url (str): Repository URL
            purl (PackageURL): Purl of component to operate on

        Returns:
            Path: Filesystem location for downloaded image
        """
        version = re.sub(pattern=r"^(sha256:)?([a-f0-9]{64})$", repl=r"sha256:\2", string=purl.version)

        version = urllib.parse.quote_plus(re.sub(r"^https?://", "", version))
        target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)

        return target_dir / f"{purl.name}@{version}"

    def _inspect_image(self, source_image: RepositoryUrl, purl: PackageURL) -> Result:
        """Verify provided image tag digest matches PURL version.

        Args:
            source_image (RepositoryUrl): Full image name
            purl (PackageURL): Purl of component to operate on

        Returns:
            Result: The result of the verification
        """
        image_prefix, _ = source_image.url.rsplit("/", maxsplit=1)
        image_id = f"{purl.name}:{purl.qualifiers.get('tag')}"

        command = [self.required_commands[0], "inspect", "--format", "{{.Digest}}", f"{image_prefix}/{image_id}"]
        inspect_command = self.run_command(command)

        try:
            inspect_command.check_returncode()
            sha_tag = inspect_command.stdout.decode().strip()
        except CalledProcessError:
            return Result.retry(message=f"Failed to get image digest for '{source_image}'")

        if sha_tag != purl.version:
            return Result.fail(
                message=f"Provided tag '{purl.qualifiers.get('tag')}' image digest does not match '{purl.version}'"
            )

        return Result.success()

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        """Copy a component to the local collection directory structure."""
        purl = hoppr.utils.get_package_url(comp.purl)

        try:
            source_image = self._get_image(url=repo_url, purl=purl)
        except HopprPluginError as ex:
            self._logger.error("%s", ex)
            self._logger.debug("purl.type:       %s", purl.type, indent_level=2)
            self._logger.debug("purl.name:       %s", purl.name, indent_level=2)
            self._logger.debug("purl.namespace:  %s", purl.namespace, indent_level=2)
            self._logger.debug("purl.version:    %s", purl.version, indent_level=2)
            self._logger.debug("purl.qualifiers: %s", purl.qualifiers, indent_level=2)

            return Result.fail(return_obj=comp, message=f"{ex}")

        target_path = self._get_target_path(repo_url, purl)
        package_out_path = target_path.parent / f"{purl.name}@{purl.version}"

        if (
            purl.type == "oci"
            and "tag" in purl.qualifiers
            and not (inspect_result := self._inspect_image(source_image, purl)).is_success()
        ):
            return inspect_result

        self._logger.info(msg=f"Copying {purl.type} image:", indent_level=2)
        self._logger.info(msg=f"source: {source_image}", indent_level=3)
        self._logger.info(msg=f"destination: {target_path}", indent_level=3)

        try:
            self._download_package(purl, repo_url, source_image.url, target_path, creds)
            self._get_package_signatures(repo_url, purl, package_out_path, creds)

        except HopprPluginError as ex:
            self._logger.info(msg="Artifact collection failed, deleting file and retrying", indent_level=2)
            target_path.unlink(missing_ok=True)
            return Result.retry(message=str(ex))

        self.set_collection_params(comp, repo_url, package_out_path)

        if Path(f"{package_out_path}.json").is_file():
            comp.properties.remove(Property(name=BomProps.COLLECTION_ARTIFACT_FILE, value=Path(package_out_path).name))
            comp.properties.append(
                Property(name=BomProps.COLLECTION_ARTIFACT_FILE, value=str(f"{Path(package_out_path).name}.json"))
            )

        return Result.success(return_obj=comp)



"""Class to store results of processes."""

from __future__ import annotations

from enum import IntEnum
from typing import TYPE_CHECKING

from typing_extensions import Self

from hoppr.exceptions import HopprError
from hoppr.models.sbom import Component, Sbom

if TYPE_CHECKING:
    import requests


class ResultStatus(IntEnum):
    """Enumeration of possible result states."""

    EXCLUDED = 0
    SUCCESS = 1
    RETRY = 2
    WARN = 3
    FAIL = 4
    SKIP = 5

    def __str__(self) -> str:
        return self.name


class Result:
    """Class to store results of processes."""

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Result) and all([
            self.status == other.status,
            self.message == other.message,
            self.return_obj == other.return_obj,
        ])

    def __init__(self, status: ResultStatus, message: str = "", return_obj: Component | Sbom | None = None):
        self.status = status
        self.message = message
        self.return_obj = return_obj

    def __str__(self) -> str:
        result_msg = f"{self.status.name}"
        if self.message != "":
            result_msg += f", msg: {self.message}"

        return result_msg.rstrip()

    @classmethod
    def success(cls, message: str = "", return_obj: Component | Sbom | None = None) -> Self:
        """Convenience method for generating success messages."""
        return cls(ResultStatus.SUCCESS, message, return_obj)

    @classmethod
    def retry(cls, message: str = "", return_obj: Component | Sbom | None = None) -> Self:
        """Convenience method for generating retry messages."""
        return cls(ResultStatus.RETRY, message, return_obj)

    @classmethod
    def fail(cls, message: str = "", return_obj: Component | Sbom | None = None) -> Self:
        """Convenience method for generating failure messages."""
        return cls(ResultStatus.FAIL, message, return_obj)

    @classmethod
    def warn(cls, message: str = "", return_obj: Component | Sbom | None = None) -> Self:
        """Convenience method for generating warning messages."""
        return cls(ResultStatus.WARN, message, return_obj)

    @classmethod
    def excluded(cls, message: str = "", return_obj: Component | Sbom | None = None) -> Self:
        """Convenience method for generating excluded messages."""
        return cls(ResultStatus.EXCLUDED, message, return_obj)

    @classmethod
    def skip(cls, message: str = "", return_obj: Component | Sbom | None = None) -> Self:
        """Convenience method for generating skip messages."""
        return cls(ResultStatus.SKIP, message, return_obj)

    def is_success(self) -> bool:
        """Convenience method for testing for success messages."""
        return self.status == ResultStatus.SUCCESS

    def is_retry(self) -> bool:
        """Convenience method for testing for retry messages."""
        return self.status == ResultStatus.RETRY

    def is_fail(self) -> bool:
        """Convenience method for testing for failure messages."""
        return self.status == ResultStatus.FAIL

    def is_warn(self) -> bool:
        """Convenience method for testing for warning messages."""
        return self.status == ResultStatus.WARN

    def is_excluded(self) -> bool:
        """Convenience method for testing for excluded messages."""
        return self.status == ResultStatus.EXCLUDED

    def is_skip(self) -> bool:
        """Convenience method for testing for skip messages."""
        return self.status == ResultStatus.SKIP

    def merge(self, other: Result):
        """Logically combine two Result objects."""
        if other.is_skip():
            return

        self.status = max(self.status, other.status)
        self.message = "\n".join(filter(None, [self.message, other.message]))

        if self.return_obj and other.return_obj:
            raise HopprError("Unable to merge two results with return objects")

        self.return_obj = next((obj for obj in [self.return_obj, other.return_obj] if obj), None)

    @staticmethod
    def from_http_response(response: requests.Response, return_obj: Component | Sbom | None = None) -> Result:
        """Build a Result object from an HTTP request response."""
        match response.status_code:
            case response_code if 200 <= response_code <= 299:
                return Result.success(
                    f"HTTP Status Code: {response.status_code}",
                    return_obj=return_obj,
                )
            case response_code if response_code >= 500:
                status = ResultStatus.RETRY
            case _:
                status = ResultStatus.FAIL

        message = f"HTTP Status Code: {response.status_code}; {response.reason or response.text}"
        return Result(status, message, return_obj=return_obj)



"""Hoppr Network utility functions."""

from __future__ import annotations

import hashlib

from pathlib import Path
from typing import TYPE_CHECKING, Final, Literal, TypeAlias

import hoppr_cyclonedx_models.cyclonedx_1_5 as cdx
import requests as requests

from pydantic import SecretStr
from requests.auth import HTTPBasicAuth

from hoppr.exceptions import HopprLoadDataError
from hoppr.models.credentials import (
    CredentialRequiredService,
    Credentials as Credentials,
)
from hoppr.utils import load_string

if TYPE_CHECKING:
    from collections.abc import Mapping
    from os import PathLike

_BLOCK_SIZE: Final[int] = 65536


HashlibAlgs: TypeAlias = Literal[
    "blake2b",
    "blake2s",
    "md5",
    "sha1",
    "sha224",
    "sha256",
    "sha384",
    "sha512",
    "sha3_224",
    "sha3_256",
    "sha3_384",
    "sha3_512",
    "shake_128",
    "shake_256",
    "blake2b_256",
]
# Mapping CDX hash algorithms to those supported by builtin `hashlib` library.
HASH_ALG_MAP: Final[Mapping[cdx.HashAlg, HashlibAlgs]] = {
    cdx.HashAlg.MD5: "md5",
    cdx.HashAlg.SHA_1: "sha1",
    cdx.HashAlg.SHA_256: "sha256",
    cdx.HashAlg.SHA_384: "sha384",
    cdx.HashAlg.SHA_512: "sha512",
    cdx.HashAlg.SHA3_256: "sha3_256",
    cdx.HashAlg.SHA3_384: "sha3_384",
    cdx.HashAlg.SHA3_512: "sha3_512",
    cdx.HashAlg.BLAKE2B_256: "blake2b_256",
    cdx.HashAlg.BLAKE2B_512: "blake2b",
}


def get_file_hash(artifact: str | PathLike[str], algorithm: cdx.HashAlg | HashlibAlgs = "sha1") -> str:
    """Compute hash of downloaded component.

    Args:
        artifact: Path to downloaded file
        algorithm: Hashing algorithm to use. Defaults to SHA1.
            The only supported `cdx.HashAlg` values are those with a matching `HashlibAlgs` in `HASH_ALG_MAP`.

    Returns:
        The computed hexidecimal hash digest.

    Raises:
        ValueError: If `algorithm` is an unsupported value of `cdx.HashAlg`.
    """
    if isinstance(algorithm, cdx.HashAlg):
        try:
            algorithm = HASH_ALG_MAP[algorithm]
        except KeyError as ex:
            raise ValueError(
                f"Hash generation using the specified algorithm ({algorithm}) is not supported by Hoppr."
            ) from ex

    artifact = Path(artifact)
    hash_obj: hashlib._Hash
    match algorithm:
        case "blake2b_256":
            hash_obj = hashlib.blake2b(artifact.read_bytes(), digest_size=32)
        case _:
            hash_obj = hashlib.new(name=algorithm)

            with artifact.open(mode="rb") as hash_bytes:
                while file_bytes := hash_bytes.read(_BLOCK_SIZE):
                    hash_obj.update(file_bytes)

    return hash_obj.hexdigest().lower()


def load_url(url: str) -> dict | list | None:
    """Load config content (either json or yml) from a url into a dict."""
    creds = Credentials.find(url)

    response = None
    if creds is not None and isinstance(creds.password, SecretStr):
        authorization_headers = {
            "PRIVATE-TOKEN": creds.password.get_secret_value(),
            "Authorization": f"Bearer {creds.password.get_secret_value()}",
        }

        basic_auth = HTTPBasicAuth(username=creds.username, password=creds.password.get_secret_value())
        response = requests.get(url, auth=basic_auth, headers=authorization_headers, timeout=60)
    else:
        response = requests.get(url, timeout=60)

    response.raise_for_status()
    valid_data = True
    try:
        if isinstance(response.content, bytes):
            return load_string(response.content.decode("utf-8"))
        if isinstance(response.content, str):
            return load_string(response.content)
        valid_data = False
    except HopprLoadDataError as parse_error:
        message = f"Unable to parse result from {url}."
        if response.url != url:
            message += f" Request was redirected to {response.url}. An auth issue might have occurred."
        raise HopprLoadDataError(message) from parse_error

    if not valid_data:
        raise HopprLoadDataError("Response type is not bytes or str")

    return None  # pragma: no cover


def download_file(
    url: str,
    dest: str,
    creds: CredentialRequiredService | None = None,
    proxies: dict[str, str] | None = None,
    timeout: int | None = 60,
    cert: tuple[str, str] | None = None,
) -> requests.Response:
    """Download content from a url into a file."""
    if creds is None:
        creds = Credentials.find(url)

    basic_auth = None
    if creds is not None and isinstance(creds.password, SecretStr):
        basic_auth = HTTPBasicAuth(username=creds.username, password=creds.password.get_secret_value())

    response = requests.get(
        url=url,
        auth=basic_auth,
        allow_redirects=True,
        proxies=proxies,
        stream=True,
        verify=True,
        timeout=timeout,
        cert=cert,
    )

    if 200 <= response.status_code < 300:
        with Path(dest).open(mode="wb") as out_file:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    out_file.write(chunk)

    return response



"""Collector plugin for DNF packages."""

from __future__ import annotations

import contextlib
import fnmatch
import itertools
import os
import shutil
import time

from configparser import ConfigParser
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import quote_plus, urlparse

import requests

from pydantic import Field, SecretStr
from requests.auth import HTTPBasicAuth

import hoppr.net
import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import BatchCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_process, hoppr_rerunner
from hoppr.exceptions import HopprPluginError
from hoppr.models.base import HopprBaseModel
from hoppr.models.credentials import Credentials
from hoppr.models.types import PurlType, RepositoryUrl
from hoppr.plugins.collect.rpm import (
    RepoConfig as RepoConfig,
    RepoConfigList,
)
from hoppr.result import Result

if TYPE_CHECKING:
    from collections.abc import Generator

    from packageurl import PackageURL

    from hoppr.models import HopprContext
    from hoppr.models.sbom import Component


def _artifact_string(purl: PackageURL) -> str:
    arch = purl.qualifiers.get("arch")
    return f"{purl.name}{f'-{purl.version}' if purl.version else ''}{f'.{arch}' if arch else ''}"


def _clear_cache(search_root: Path):
    """Clear the cache for a given Path."""
    temp_paths = list(search_root.rglob("hoppr-tmp-*"))
    for folder in temp_paths:
        if folder.is_dir():
            shutil.rmtree(folder, ignore_errors=True)
        else:
            folder.unlink()


def _is_rpm_repo(repo_url: str) -> bool:
    url = RepositoryUrl(url=repo_url) / "repodata/repomd.xml"
    basic_auth = None
    creds = Credentials.find(repo_url)
    if creds is not None and isinstance(creds.password, SecretStr):
        basic_auth = HTTPBasicAuth(username=creds.username, password=creds.password.get_secret_value())
    for attempt in range(3):
        if attempt > 0:
            time.sleep(5)

        resp = requests.get(f"{url}", auth=basic_auth, allow_redirects=False, stream=True, verify=True, timeout=60)
        if resp.status_code < 300:
            return True
        if resp.status_code < 500:
            return False

    return False


def _repo_proxy(url: str) -> str:
    proxy = os.getenv("https_proxy", "_none_")
    no_proxy_urls = [item for item in os.getenv("no_proxy", "").split(",") if item != ""]

    parsed_url = urlparse(url)

    for pattern in no_proxy_urls:
        if pattern in parsed_url.netloc or fnmatch.fnmatch(name=parsed_url.netloc, pat=pattern):
            proxy = "_none_"
            break

    return proxy


class CollectDnfConfig(HopprBaseModel):
    """Configuration options for CollectRpmPlugin."""

    dnf_command: str = Field(default="dnf", description="Command to use for DNF collection")
    base_arch: str | None = Field(default=None, description="Value for $basearch in repository configs")
    release_ver: str | None = Field(default=None, description="Value for $releasever in repository configs")
    vars_dir: Path | None = Field(
        default=next((Path("/", "etc", name, "vars") for name in ["dnf", "yum"]), None),
        description="DNF variable definitions directory.",
    )


class CollectDnfPlugin(
    BatchCollectorPlugin,
    required_commands=["dnf"],
    supported_purl_types=["rpm"],
    products=["rpm/*"],
):
    """Collector plugin for DNF packages."""

    def __init__(
        self,
        context: HopprContext,
        config: dict | None = None,
        config_file: str | PathLike = Path.cwd() / ".hoppr-dnf" / "dnf.conf",
    ) -> None:
        super().__init__(context=context, config=config)

        self.collector_config = CollectDnfConfig(**(config or {}))
        self.repo_configs = RepoConfigList()
        self.manifest_repos: list[str] = [f"{repo.url}" for repo in self.context.repositories[PurlType.RPM]]

        self.password_list: list[str] = []
        self.config_file = Path(config_file)

        self.required_commands = [self.collector_config.dnf_command.split()[0]]

        self.base_command = [
            *self.collector_config.dnf_command.split(),
            "--quiet",
            "--disableexcludes=all",
            f"--config={self.config_file}",
            "--disablerepo=*",
            "--enablerepo=hoppr-tmp-*",
        ]

        self.debug_args = ["--verbose"]

    def _get_cert(self, repo_url: str) -> tuple[str, str] | None:
        """Get the cert for a local repository."""
        cert_repos = [
            repo
            for repo in self.repo_configs
            if repo.ssl_client_cert and repo.ssl_client_key
            and any(url for url in repo.base_url if url in repo_url)
        ]  # fmt: skip

        return next(((str(cert_repo.ssl_client_cert), str(cert_repo.ssl_client_key)) for cert_repo in cert_repos), None)

    def _get_component_download_url(self, purl: PackageURL) -> str:
        artifact = _artifact_string(purl)

        command = [*self.base_command, "repoquery", "--location", artifact]
        run_result = self.run_command(command, self.password_list)

        # If RPM URL not found, no need to try downloading it
        if run_result.returncode != 0 or len(run_result.stdout.decode("utf-8")) == 0:
            msg = f"{self.required_commands[0]} failed to locate package for {purl}"
            self.get_logger().debug(msg=msg, indent_level=2)

            raise HopprPluginError(msg)

        # Taking the first URL if multiple are returned
        return run_result.stdout.decode("utf-8").strip().split("\n")[0]

    def _get_found_repo(self, found_url: str) -> str | None:
        """Identify the repository associated with the specified URL."""
        # Collect manifest repos
        repo_urls = self.manifest_repos

        # Collect local repos
        repo_urls.extend(itertools.chain.from_iterable(repo.base_url for repo in self.repo_configs))

        return next((repo for repo in repo_urls if repo in found_url), None)

    def _get_repo_files(self) -> Generator[Path, None, None] | None:
        """Return a list of the system repo files and populate the plugin repo config."""
        repo_files = None

        # Get all system-managed .repo files
        repo_files = Path("/", "etc", "yum.repos.d").glob("*.repo")

        self._populate_config_list()

        return repo_files

    def _populate_config_list(self):
        """Populate the repo config list with parsed repo files."""
        # Set RepoConfig class variables
        RepoConfig.base_arch = self.collector_config.base_arch
        RepoConfig.release_ver = self.collector_config.release_ver
        RepoConfig.vars_dir = self.collector_config.vars_dir

        repo_files = Path("/", "etc", "yum.repos.d").glob("*.repo")

        for file in repo_files:
            self.repo_configs.extend(RepoConfigList.parse_file(file))

    def get_version(self) -> str:  # noqa: D102
        return __version__

    @hoppr_process
    def pre_stage_process(self) -> Result:  # noqa: D102
        repo_config = ConfigParser()
        repo_config["main"] = {"cachedir": f"{self.config_file.parent / 'cache'}"}

        # Clear out cache before starting the pre_stage_process
        search_root = self.config_file.parent / "cache"
        _clear_cache(search_root)

        for idx, repo in enumerate(self.context.repositories[PurlType.RPM]):
            temp_repo = f"hoppr-tmp-{idx}"
            creds = Credentials.find(f"{repo.url}")
            if not _is_rpm_repo(f"{repo.url}"):
                self.get_logger().warning(
                    "Repo %s is not an RPM repository (repomd.xml file not found), will not be searched", repo.url
                )
                continue

            self.get_logger().debug("Creating repo %s for url %s", temp_repo, repo.url)

            # Create temporary repository file
            repo_config[temp_repo] = {
                "baseurl": f"{repo.url}",
                "enabled": "1",
                "name": f"Hoppr temp repository {idx}",
                "priority": "1",
                "proxy": _repo_proxy(f"{repo.url}"),
                "module_hotfixes": "true",
            }

            if creds is not None and isinstance(creds.password, SecretStr):
                repo_config[temp_repo]["username"] = f"{creds.username}"
                repo_config[temp_repo]["password"] = f"{creds.password.get_secret_value()}"

            self.repo_configs.append(RepoConfig.parse_obj(repo_config[temp_repo]))

        repo_files = self._get_repo_files()

        if not self.context.strict_repos and repo_files:
            # Get all system-managed .repo files
            system_repos = ConfigParser()
            system_repos.read([str(repo_file) for repo_file in repo_files])

            # Get enabled system repos and add to temporary repository file
            for section in system_repos.sections():
                if bool(system_repos[section]["enabled"]) and system_repos[section]["enabled"] != "false":
                    temp_repo = f"hoppr-tmp-{section}"

                    with contextlib.suppress(KeyError):
                        repo_url = system_repos[section]["baseurl"]
                        self.get_logger().debug(msg=f"Creating repo {temp_repo} for url {repo_url}")
                        repo_config.add_section(temp_repo)
                        repo_config[temp_repo] = dict(system_repos[section])
        try:
            # Create repo config dir in user directory
            config_dir = self.config_file.parent
            config_dir.mkdir(parents=True, exist_ok=True)
            with self.config_file.open(mode="w+", encoding="utf-8") as repo_file:
                repo_config.write(repo_file, space_around_delimiters=False)
        except OSError as ex:
            return Result.fail(f"Unable to write DNF repository config file: {ex}")

        # Populate user DNF cache
        command = [
            *self.base_command,
            "check-update",
            "makecache",
            *(self.debug_args if self.get_logger().is_verbose() else []),
        ]

        # Generate cache to use when downloading components
        result = self.run_command(command, self.password_list)
        self.get_logger().debug(msg=(f"Return Code: {result.returncode}"), indent_level=2)
        if result.returncode != 0:
            if hoppr.utils.get_partition(config_dir).fstype == "vboxsf":
                self.get_logger().warning(
                    "NOTE: Possible attempt to run DNF from a shared vagrant directory('%s'). "
                    "Recommend retrying outside of the shared vagrant directory",
                    config_dir,
                )
            return Result.fail(message="Failed to populate DNF cache.")

        return Result.success()

    @hoppr_rerunner
    def collect(self, comp: Component) -> Result:
        """Copy a component to the local collection directory structure."""
        purl = hoppr.utils.get_package_url(comp.purl)

        self.get_logger().info(msg=f"Copying DNF package from {purl}", indent_level=2)

        # Try getting RPM URL
        try:
            found_url = self._get_component_download_url(purl)
        except HopprPluginError as ex:
            return Result.retry(message=str(ex))

        self._populate_config_list()
        repo = self._get_found_repo(found_url)

        # Return failure if found RPM URL is not from a repo defined in the manifest
        if repo is None:
            return Result.fail(
                "Successfully found RPM file but URL does not match any repository in manifest."
                f" (Found URL: '{found_url}')"
            )

        result = self.check_purl_specified_url(purl, repo)
        if not result.is_success():
            return result

        found_url_path = Path(found_url)
        subdir = found_url_path.relative_to(repo).parent
        target_dir = self.directory_for(purl.type, repo, subdir=str(subdir))
        package_out_path = target_dir / f"{purl.name}-{purl.version}.rpm"

        # Download the RPM file to the new directory
        response = hoppr.net.download_file(
            url=found_url,
            dest=str(target_dir / found_url_path.name),
            creds=Credentials.find(url=repo),
            cert=self._get_cert(repo),
            timeout=self.process_timeout,
        )
        result = Result.from_http_response(response)

        if result.is_retry():
            msg = f"Failed to download DNF artifact {purl.name} version {purl.version} from {found_url}"
            self.get_logger().error(msg=msg, indent_level=2)
            return Result.retry(message=msg)

        if result.is_success():
            self.set_collection_params(comp, repo, package_out_path)
            result = Result.success(result.message, return_obj=comp)

        return result

    @hoppr_process
    def post_stage_process(self) -> Result:  # noqa: D102
        # Find repodata folders created when cache was generated
        search_root = self.config_file.parent / "cache"
        paths = list(search_root.rglob("hoppr-tmp-*/repodata"))

        repo_config = ConfigParser()
        repo_config.read(filenames=self.config_file, encoding="utf-8")

        msg = "Copying DNF repodata into bundle"
        self.get_logger().info(msg)
        self.notify(msg, type(self).__name__)

        # Loop over repos defined in the temp dnf.conf file
        for section in repo_config.sections():
            # Loop over repodata folders in DNF cache
            for folder in paths:
                # Check if full repodata path contains temp repo name
                if section in str(folder):
                    with contextlib.suppress(KeyError):
                        # Get temp repo's URL as defined in manifest
                        repo_url = repo_config[section]["baseurl"]
                        target_dir = self.context.collect_root_dir / "rpm" / quote_plus(repo_url) / "repodata"

                        shutil.copytree(src=folder, dst=target_dir, dirs_exist_ok=True)

        msg = f"Removing config file {self.config_file}"
        self.get_logger().info(msg)
        self.notify(msg, type(self).__name__)

        if self.config_file.exists():
            try:
                self.config_file.unlink()
            except FileNotFoundError as ex:
                return Result.fail(f"Failed to remove temporary DNF config file: {ex}")

        msg = "Removing cached DNF repodata"
        self.get_logger().info(msg)
        self.notify(msg, type(self).__name__)

        # Clear cache at the end of the post stage process
        _clear_cache(search_root)

        return Result.success()



"""Models to express `Annotations.annotator` fields."""

from __future__ import annotations

from typing import Annotated, ClassVar

import hoppr_cyclonedx_models.cyclonedx_1_5 as cdx

from pydantic import Field

from hoppr.models.base import CycloneDXBaseModel, UniqueIDMap


class AnnotatorBase(CycloneDXBaseModel):
    """Annotations data model derived from CycloneDXBaseModel."""

    organization: Annotated[
        cdx.OrganizationalEntity | None,
        Field(
            description="The organization that created the annotation",
        ),
    ] = None
    individual: Annotated[
        cdx.OrganizationalContact | None,
        Field(
            description="The person that created the annotation",
        ),
    ] = None
    component: Annotated[
        cdx.Component | None,
        Field(
            description="The tool or component that created the annotation",
        ),
    ] = None
    service: Annotated[
        cdx.Service | None,
        Field(
            description="The service that created the annotation",
        ),
    ] = None


class AnnotatorOrganizationRequired(AnnotatorBase):
    """Annotations.annotator item model with required `organization` field."""

    organization: cdx.OrganizationalEntity

    unique_id_map: ClassVar[UniqueIDMap] = {}


class AnnotatorIndividualRequired(AnnotatorBase):
    """Annotations.annotator item model with required `individual` field."""

    individual: cdx.OrganizationalContact

    unique_id_map: ClassVar[UniqueIDMap] = {}


class AnnotatorComponentRequired(AnnotatorBase):
    """Annotations.annotator item model with required `component` field."""

    component: cdx.Component

    unique_id_map: ClassVar[UniqueIDMap] = {}


class AnnotatorServiceRequired(AnnotatorBase):
    """Annotations.annotator item model with required `service` field."""

    service: cdx.Service

    unique_id_map: ClassVar[UniqueIDMap] = {}


Annotator = Annotated[
    AnnotatorOrganizationRequired
    | AnnotatorIndividualRequired
    | AnnotatorComponentRequired
    | AnnotatorServiceRequired,
    Field(
        description="The organization, person, component, or service which "
        "created the textual content of the annotation.",
        title="Annotator",
    ),
]  # fmt: skip


class Annotations(CycloneDXBaseModel, cdx.Annotations):
    """Annotations data model derived from CycloneDXBaseModel."""

    def __hash__(self) -> int:
        return hash(self.bom_ref)

    annotator: Annotator

    unique_id_map: ClassVar[UniqueIDMap] = {}



"""Hoppr exceptions and warnings."""

from __future__ import annotations


class HopprError(RuntimeError):
    """Base exception raised within the hoppr app."""


class HopprPluginError(HopprError):
    """Exception raised for errors working with plug-ins."""


class HopprPluginRetriableError(HopprPluginError):
    """Exception raised when a plugin fails to perform an action that can be retried."""


class HopprLoadDataError(HopprError):
    """Exception raised for errors loading json/yml data."""


class HopprCredentialsError(HopprError):
    """Exception raised for errors loading credential data."""


class HopprPrivateKeyError(HopprError):
    """Exception raised for unsupported private key usage."""


class HopprValidationError(ValueError):  # pragma: no cover
    """Exception raised for failures during SBOM validation."""

    def __init__(self, *args: object, check_name: str) -> None:
        super().__init__(*args)
        self.check_name = check_name

    def __str__(self) -> str:
        return "".join(str(arg) for arg in self.args)


class HopprExperimentalWarning(UserWarning):
    """Warning raised when experimental features are accessed."""



"""A logger that gets dumped to stdout when closed."""

from __future__ import annotations

import functools
import inspect
import itertools
import logging

from collections.abc import Callable, MutableMapping
from logging import FileHandler, Formatter, Logger, LoggerAdapter
from logging.handlers import MemoryHandler
from typing import TYPE_CHECKING, cast

from typing_extensions import override

if TYPE_CHECKING:
    from threading import _RLock as RLock


def locked(func: Callable) -> Callable:
    """Acquire logfile lock, run the wrapped function, then release the lock."""

    @functools.wraps(func)
    def wrapper(self: HopprLogger, *args, **kwargs):
        if self.lock:
            self.lock.acquire()

        func(self, *args, **kwargs)

        if self.lock:
            self.lock.release()

    return wrapper


class HopprLogger(LoggerAdapter):
    """Logger that buffers log records in memory until flushed."""

    id_iter = itertools.count()

    def __init__(
        self,
        name: str | None = None,
        filename: str = "hoppr.log",
        level: int = logging.INFO,
        lock: RLock | None = None,
        flush_immed: bool = False,
    ) -> None:
        """Initialize the logger, formatter, and handlers.

        Args:
            name (str | None, optional): Name to assign the logger. Defaults to None.
            filename (str, optional): Log file to create. Defaults to "hoppr.log".
            level (int, optional): Logging level. Defaults to logging.INFO.
            lock (RLock | None, optional): Lock object to acquire/release when writing to log file. Defaults to None.
            flush_immed (bool, optional): Flush immediately when logging. Defaults to False.
        """
        self.extra: MutableMapping[str, object]
        self.logger: Logger

        self.filename = filename
        self.flush_immed: bool = flush_immed
        self.instance_id: int = next(self.id_iter)
        self.lock: RLock | None = lock

        _, frame_info, *_ = inspect.stack(context=2)
        caller = frame_info.function

        if (caller_cls := frame_info.frame.f_locals.get("self")) is not None:
            caller = type(caller_cls).__name__

        extra = cast(MutableMapping, {})
        extra["caller"] = name or caller

        logger = logging.getLogger(f"{name}-{self.instance_id}" if name else f"{caller}-{self.instance_id}")
        logger.setLevel(level)

        formatter = Formatter(
            fmt="[$asctime] - [${caller}] - [$levelname] - $message",
            style="$",
            defaults={"caller": caller},
        )

        file_handler = FileHandler(filename)
        file_handler.setFormatter(formatter)

        log_handler = MemoryHandler(10000, flushLevel=logging.CRITICAL, target=file_handler)
        logger.addHandler(log_handler)

        super().__init__(logger, extra)

    def clear_targets(self) -> None:
        """Set target for all MemoryHandlers in this logger to None.

        Thus when these handlers are flushed, nothing will go to standard output
        """
        for handler in self.logger.handlers:
            if isinstance(handler, MemoryHandler):
                handler.setTarget(None)

    @locked
    def close(self) -> None:
        """Close (and flush) all handlers for this logger."""
        for handler in self.logger.handlers:
            handler.close()

    @locked
    def flush(self) -> None:
        """Flush all handlers for this logger."""
        for handler in self.logger.handlers:
            handler.flush()

    def is_verbose(self) -> bool:
        """Check whether running with `--debug`/`--verbose` flag."""
        return self.logger.level == logging.DEBUG

    def log(self, level: int, msg: object, *args, **kwargs) -> None:
        """Wrapper function for logging messages."""
        super().log(level, msg, *args, **kwargs)

        if self.flush_immed:
            self.flush()

    @override
    def process(self, msg: str, kwargs: MutableMapping[str, object]) -> tuple[str, MutableMapping[str, object]]:
        indent_level = cast(int, kwargs.pop("indent_level", 0))
        indent_string = " " * 4 * indent_level
        msg = f"{indent_string}{msg}"

        return super().process(msg, kwargs)



"""Models to describe a Code Climate issue report."""

from __future__ import annotations

import hashlib

from enum import Enum
from pathlib import Path  # noqa: TC003
from typing import TYPE_CHECKING, Annotated, Any, Literal

from pydantic import Field, validator

from hoppr.models.base import HopprBaseModel

if TYPE_CHECKING:
    from collections.abc import Iterable

    from rich.repr import RichReprResult


def _normalize_path(path: Path) -> str:
    return str(path).replace("\\", "/")


class IssueCategory(str, Enum):
    """Category indicating the nature of the issue being reported."""

    BUG_RISK = "Bug Risk"
    CLARITY = "Clarity"
    COMPATIBILITY = "Compatibility"
    COMPLEXITY = "Complexity"
    DUPLICATION = "Duplication"
    PERFORMANCE = "Performance"
    SECURITY = "Security"
    STYLE = "Style"


class IssueSeverity(str, Enum):
    """Severity of the issue."""

    INFO = "info"
    MINOR = "minor"
    MAJOR = "major"
    CRITICAL = "critical"
    BLOCKER = "blocker"


class IssueContent(HopprBaseModel):
    """A markdown snippet describing the issue, including deeper explanations and links to other resources."""

    body: Annotated[str | None, Field(description="Body of the snippet, formatted as markdown.")]


class LineRange(HopprBaseModel):
    """Beginning and end line number to define a range of lines."""

    begin: int
    end: int


class PositionReference(HopprBaseModel):
    """Reference to the a specific character position in a file."""

    line: int
    column: int


class PositionRange(HopprBaseModel):
    """Beginning and end position to define a range of characters."""

    begin: PositionReference
    end: PositionReference


class IssueLocationLineRange(HopprBaseModel):
    """Line-based locations.

    Specify a beginning and end line number for the issue location, which form a range.
    Line numbers are 1-based, so the first line of a file would be represented by 1.
    """

    path: Annotated[Path, Field(description="File path relative to working directory.")]
    lines: LineRange

    _validate_path = validator("path", always=True, allow_reuse=True)(_normalize_path)


class IssueLocationPositionRange(HopprBaseModel):
    """Position-based location.

    Allows more precision by including references to the specific characters
    that form the source code range representing the issue.
    """

    path: Annotated[Path, Field(description="File path relative to working directory.")]
    positions: PositionRange

    _validate_path = validator("path", always=True, allow_reuse=True)(_normalize_path)


IssueLocation = IssueLocationLineRange | IssueLocationPositionRange


class IssueTrace(HopprBaseModel):
    """Represents ordered or unordered lists of source code locations."""

    locations: list[IssueLocation]
    trace: bool = False


class Issue(HopprBaseModel):
    """Represents a single instance of a validation check finding."""

    type: Annotated[Literal["issue"], Field(description='Must always be "issue".')] = "issue"
    check_name: Annotated[
        str | None,
        Field(description="A unique name representing the static analysis check that emitted this issue."),
    ] = None
    description: Annotated[str, Field(description="A string explaining the issue that was detected.")]
    content: Annotated[
        str | None,
        Field(
            description=(
                "A markdown snippet describing the issue, including deeper explanations and links to other resources."
            ),
        ),
    ] = None
    categories: Annotated[
        list[IssueCategory] | None,
        Field(description="At least one category indicating the nature of the issue being reported."),
    ] = None
    location: Annotated[
        IssueLocation,
        Field(
            description="A `Location` object representing the place in the source code where the issue was discovered."
        ),
    ]
    other_locations: list[IssueLocation] | None = None
    trace: Annotated[
        IssueTrace | None,
        Field(
            description="A `Trace` object representing other interesting source code locations related to this issue."
        ),
    ] = None
    remediation_points: Annotated[
        int | None,
        Field(
            description=(
                "An integer indicating a rough estimate of how long it would take to resolve the reported issue."
            )
        ),
    ] = None
    severity: Annotated[
        IssueSeverity,
        Field(
            description=(
                "A `Severity` string (`info`, `minor`, `major`, `critical`, or `blocker`) "
                "describing the potential impact of the issue found."
            ),
        ),
    ]
    fingerprint: Annotated[
        str,
        Field(
            description=(
                "A unique, deterministic identifier for the specific issue being "
                "reported to allow a user to exclude it from future analyses."
            ),
            regex="[a-z0-9]{13,64}",
        ),
    ] = None  # type: ignore[assignment]

    @validator("fingerprint", allow_reuse=True, pre=True, always=True)
    @classmethod
    def _validate_fingerprint(cls, fingerprint: str | None, values: dict[str, Any]) -> str:
        if fingerprint:
            return fingerprint

        hash_ = hashlib.shake_256(str(values).encode(encoding="utf-8"))
        return hash_.hexdigest(8)

    def __hash__(self) -> int:
        return int.from_bytes(self.fingerprint.encode(encoding="utf-8"), byteorder="big")


class IssueList(HopprBaseModel):
    """List of Issue objects."""

    __root__: list[Issue]

    def __getitem__(self, idx: int) -> Issue:
        return self.__root__[idx]

    def __init__(self, issues: Iterable[Issue] | None = None, **data):
        data["__root__"] = data.get("__root__", list(issues or []))
        super().__init__(**data)

    def __iter__(self):
        yield from self.__root__

    def __len__(self) -> int:
        return len(self.__root__)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__root__})"

    def __rich_repr__(self) -> RichReprResult:
        yield self.__root__

    def append(self, item: Issue) -> None:
        """Append object to the end of the list."""
        self.__root__.append(item)  # pragma: no cover

    def extend(self, iterable: Iterable[Issue]) -> None:
        """Extend list by appending elements from the iterable."""
        self.__root__.extend(iterable)  # pragma: no cover


IssueLocationPositionRange.update_forward_refs()



"""Define constants used by Hoppr."""

from __future__ import annotations

from enum import Enum


class BomProps(str, Enum):
    """SBOM Property Field Names."""

    COLLECTION_DIRECTORY = "hoppr:collection:directory"
    COLLECTION_PLUGIN = "hoppr:collection:plugin"
    COLLECTION_REPOSITORY = "hoppr:collection:repository"
    COLLECTION_TIMETAG = "hoppr:collection:timetag"
    COLLECTION_ARTIFACT_FILE = "hoppr:collection:artifact_file"

    COMPONENT_SEARCH_SEQUENCE = "hoppr:repository:component_search_sequence"


class ConfigKeys(str, Enum):
    """Keys used in configuration files."""

    CONFIG = "config"
    PLUGINS = "plugins"



"""Configuration file models for `hopctl validate sbom`."""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any, Final, TypeAlias

from jinja2 import Environment, PackageLoader
from pydantic import Field, validator
from typing_extensions import Self

from hoppr.models.base import HopprBaseModel
from hoppr.models.validation.base import BaseCheckConfig
from hoppr.models.validation.exclude import ExcludeConfig

if TYPE_CHECKING:
    DictStrAny: TypeAlias = dict[str, Any]


class OnCheckFail(str, Enum):
    """Values to configure response behavior for individual checks."""

    IGNORE = "ignore"
    WARN = "warn"
    ERROR = "error"

    def __str__(self) -> str:
        return self.value


class LicenseCheckConfig(BaseCheckConfig):
    """Configuration for SBOM license checks."""

    # Override parent class to bypass singleton constraint
    def __new__(cls, *args, **kwargs) -> Self:  # noqa: D102
        return super(HopprBaseModel, cls).__new__(cls)

    name_or_id: OnCheckFail = Field(
        default=OnCheckFail.WARN,
        description="One of the `license.id`, `license.name` fields must be specified",
    )
    last_renewal: OnCheckFail = Field(
        default=OnCheckFail.IGNORE,
        description="The `license.licensing.lastRenewal` field is not specified",
    )
    purchase_order: OnCheckFail = Field(
        default=OnCheckFail.IGNORE,
        description="The `license.licensing.purchaseOrder` field is not specified",
    )
    license_types: OnCheckFail = Field(
        default=OnCheckFail.WARN,
        description="The `license.licensing.licenseType` field is either not specified or empty",
    )
    licenses_field: OnCheckFail = Field(
        default=OnCheckFail.WARN,
        description="The `licenses` field is either not specified or empty",
    )
    expiration: OnCheckFail = Field(
        default=OnCheckFail.IGNORE,
        description="The `license.licensing.expiration` field contains an expired or soon-to-expire date",
    )


class ComponentCheckConfig(BaseCheckConfig):
    """Configuration for SBOM component checks."""

    licenses: LicenseCheckConfig = Field(
        default=LicenseCheckConfig(),
        description="Configurations for the `components[].licenses` checks",
    )
    name_field: OnCheckFail = Field(
        default=OnCheckFail.WARN,
        description="The `components[].name` field is not specified",
    )
    supplier_field: OnCheckFail = Field(
        default=OnCheckFail.WARN,
        description="The `components[].supplier` field is not specified",
    )
    version_field: OnCheckFail = Field(
        default=OnCheckFail.WARN,
        description="The `components[].version` field is not specified",
    )
    unique_id: OnCheckFail = Field(
        default=OnCheckFail.WARN,
        description="One of the `components[].purl`, `components[].cpe`, `components[].swid` fields must be specified",
    )


class MetadataCheckConfig(BaseCheckConfig):
    """Configuration for SBOM metadata checks."""

    licenses: LicenseCheckConfig = Field(
        default=LicenseCheckConfig(),
        description="Configurations for the `metadata.licenses` checks",
    )
    authors: OnCheckFail = Field(
        default=OnCheckFail.WARN,
        description="One of the `metadata.authors`, `metadata.tools` fields must be specified and not empty",
    )
    supplier_field: OnCheckFail = Field(
        default=OnCheckFail.WARN,
        description="The `metadata.supplier` field is not specified",
    )
    timestamp: OnCheckFail = Field(
        default=OnCheckFail.WARN,
        description="The `metadata.timestamp` field is either not specified or invalid",
    )


class SbomCheckConfig(BaseCheckConfig):
    """Configuration for top-level SBOM field checks."""

    components_field: OnCheckFail = Field(
        default=OnCheckFail.WARN,
        description="The `components` field is either not specified or empty",
    )
    spec_version: OnCheckFail = Field(
        default=OnCheckFail.WARN,
        description="The `specVersion` should be 1.5",
    )
    unique_id: OnCheckFail = Field(
        default=OnCheckFail.WARN,
        description="The `serialNumber` field is either not specified or invalid",
    )
    vulnerabilities_field: OnCheckFail = Field(
        default=OnCheckFail.WARN,
        description="The `vulnerabilities` field is either not specified or empty",
    )


_CONFIG_FILE_HEADER: Final[str] = """\
# Default values for `hopctl validate sbom` configuration options and checks.
#
# Each check listed will accept a value of either `ignore`, `warn`, or `error` to configure
# desired response behavior on failure.
#
# Items can be excluded from validation altogether by specifying a pattern under the top-level
# `exclude:` key. Any items matching these patterns will not be evaluated during validation.
#
# A pattern can be one of the following forms:
#
# - `jmespath:<JMESPath search expression>`
# - An object representing objects to exclude based on the fields and patterns specified
#
# In the latter form, the name of each first child of the `exclude:` object must be one of the
# types defined as an array by the CycloneDX specification.
#
# ```yaml
# exclude:
#   <CycloneDX array object name>:  # e.g. `components`, `licenses`
#     # multiple fields defined on the same list item represent an AND relationship
#     - <field name 1>: <field value or pattern>
#       <field name 2>: <field value or pattern>
#
#     # separate list items represent an OR relationship when filtering
#     - <field name 3>: <field value or pattern>
#
#     # alternatively, exclusion filter can be specified with a JMESPath expression
#     - jmespath:<JMESPath search expression>
#
#     ...
# ```
#
# The `<field value or pattern>` above can be one of:
#
# - `regexp:/<regular expression identifying elements to exclude>/[gmi]*`
# - literal string representing the desired exclusion value
"""  # fmt: skip


class ValidateConfig(BaseCheckConfig):
    """Model for SBOM validation configuration."""

    exclude: ExcludeConfig | None = None
    sbom_checks: SbomCheckConfig = SbomCheckConfig()
    metadata_checks: MetadataCheckConfig = MetadataCheckConfig()
    component_checks: ComponentCheckConfig = ComponentCheckConfig()

    @validator("exclude", allow_reuse=True, always=True, pre=True)
    @classmethod
    def _default_exclude(cls, exclude: DictStrAny | None) -> DictStrAny:
        return exclude or {"components": [{"type": "operating-system"}]}

    def yaml(self, *args, **kwargs) -> str:
        """YAML representation of current config settings."""
        environment = Environment(
            loader=PackageLoader(
                package_name="hoppr.models.validation",
                package_path="templates",
            )
        )

        template = environment.get_template("config.yml.jinja2")

        return template.render(fields=self.__fields__.values(), obj=self, header=_CONFIG_FILE_HEADER)


ValidateConfig.update_forward_refs()



"""Collector plugin for PyPI packages."""

from __future__ import annotations

import importlib.util
import re
import sys

from enum import Enum
from typing import TYPE_CHECKING, Literal, TypeAlias

import jmespath
import requests
import xmltodict

from pydantic import Field, SecretStr, validator
from requests.auth import HTTPBasicAuth

import hoppr.utils

from hoppr import __version__, cdx
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.constants import BomProps
from hoppr.exceptions import HopprPluginError, HopprPluginRetriableError
from hoppr.models.base import HopprBaseModel
from hoppr.models.sbom import Hash, Property
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from packageurl import PackageURL

    from hoppr.models import HopprContext
    from hoppr.models.credentials import CredentialRequiredService
    from hoppr.models.sbom import Component


class CollectionType(str, Enum):
    """Represents a collection type for the PyPI collector."""

    BINARY = "binary"
    BINARY_ONLY = "binary-only"
    BINARY_PREFERRED = "binary-preferred"
    BOTH_PREFERRED = "both-preferred"
    BOTH_REQUIRED = "both-required"
    SOURCE = "source"
    SOURCE_ONLY = "source-only"
    SOURCE_PREFERRED = "source-preferred"

    def __str__(self) -> str:
        return self.value.lower()

    @classmethod
    def _missing_(cls, value: object):  # noqa: ANN206
        return next((member for member in cls if member.name in {str(value), str(value).upper()}), None)


ConcreteCollectionType: TypeAlias = Literal["binary", "source"]


class CollectPypiConfig(HopprBaseModel, underscore_attrs_are_private=True):
    """Configuration options for CollectPypiPlugin."""

    _type_order: tuple[ConcreteCollectionType] | tuple[ConcreteCollectionType, ConcreteCollectionType]

    pip_command: list[str] = Field(default=[sys.executable, "-m", "pip"])
    type: CollectionType = Field(default=CollectionType.BINARY_PREFERRED)

    @validator("pip_command", allow_reuse=True, pre=True)
    @classmethod
    def pip_cmd_to_list(cls, pip_cmd: list[str] | str) -> list[str]:
        """Convert the pip command to a list[str] before pydantic validation."""
        return pip_cmd.split() if isinstance(pip_cmd, str) else pip_cmd

    @validator("type", allow_reuse=True, pre=True)
    @classmethod
    def type_to_lower(cls, type_: str) -> str:
        """Convert the type to lowercase before pydantic validation."""
        return type_.lower()

    def __init__(self, **data):
        super().__init__(**data)

        match self.type.lower():
            case "binary-only":
                self._type_order = ("binary",)
            case "source-only":
                self._type_order = ("source",)
            case "binary" | "binary-preferred" | "both-preferred" | "both-required":
                self._type_order = ("binary", "source")
            case "source" | "source-preferred":
                self._type_order = ("source", "binary")
            case invalid_type:  # pragma: no cover
                raise ValueError(f"Invalid PyPI collection type specified: {invalid_type}")


class CollectPypiPlugin(
    SerialCollectorPlugin,
    supported_purl_types=["pypi"],
    required_commands=[sys.executable],
    products=["pypi/*"],
    system_repositories=["https://pypi.org/simple"],
):
    """Collector plugin for PyPI packages."""

    def get_version(self) -> str:  # noqa: D102
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)

        self.collector_config = CollectPypiConfig(**(config or {}))

        self.manifest_repos: list[str] = []
        self.password_list: list[str] = []

    def _download_package(
        self, purl: PackageURL, download_url: str, target_dir: Path, creds: CredentialRequiredService | None
    ) -> str:
        """Downloads a Python package from a specific repository URL.

        Args:
            purl: Used to identify the package in the logs.
            download_url: Where the package will be downloaded from.
            target_dir: Where the package will be downloaded to.
            creds: Credentials used for the HTTP request.

        Returns:
            A success message.

        Raises:
            HopprPluginRetryException: If package download fails and should be retried.
        """
        password_list = []

        if creds is not None and isinstance(creds.password, SecretStr):
            download_url = str(
                RepositoryUrl(
                    url=download_url,
                    username=creds.username,
                    password=creds.password.get_secret_value(),
                )
            )
            password_list = [creds.password.get_secret_value()]

        command = [
            *self.collector_config.pip_command,
            "download",
            "--no-deps",
            "--no-cache",
            "--timeout",
            str(self.process_timeout),
            "--index-url",
            download_url,
            "--dest",
            str(target_dir),
            f"{purl.name}=={purl.version}",
            *(["--verbose"] if self.get_logger().is_verbose() else []),
        ]

        collection_type = self.collector_config.type

        success_count = 0
        successes_needed = 2 if collection_type == "both-required" else 1

        self.get_logger().info("Pypi collection type: %s", collection_type, indent_level=2)

        result_codes = []

        for attempt_type in self.collector_config._type_order:
            self.get_logger().info("Attempting %s collection", attempt_type, indent_level=2)

            run_result = self.run_command([*command, self._pypi_param_for_type(attempt_type)], password_list)

            result_codes.append(run_result.returncode)

            if run_result.returncode == 0:
                success_count += 1
                self.get_logger().debug("Collection of %s successful", attempt_type, indent_level=2)
            elif not collection_type.endswith("preferred"):
                self.get_logger().debug("Collection of %s failed", attempt_type, indent_level=2)
                raise HopprPluginRetriableError(self._get_retry_exception_message(success_count))

            if success_count >= successes_needed and self.collector_config.type != CollectionType.BOTH_PREFERRED:
                break

        if success_count >= successes_needed:
            return self._get_success_message(success_count, result_codes)

        raise HopprPluginRetriableError(self._get_retry_exception_message(success_count))

    def _get_download_urls(
        self,
        component: Component,
        repo_url: str,
        creds: CredentialRequiredService | None = None,
    ) -> list[dict[str, str]]:
        auth = None
        if creds is not None and isinstance(creds.password, SecretStr):
            auth = HTTPBasicAuth(creds.username, creds.password.get_secret_value())

        response = requests.get(url=f"{repo_url}/simple/{component.name}", auth=auth, timeout=60)

        xml_input = response.text.removeprefix("<!DOCTYPE html>")
        xml_input = re.sub(pattern=r"\s*<head>[.\S\s\n]*</head>", repl="", string=xml_input)
        response_dict = xmltodict.parse(xml_input=xml_input, process_comments=False)

        links: list[dict[str, str]] = jmespath.search(expression="html.body.a", data=response_dict) or []

        # Normalize filenames
        for link in links:
            link["#text"] = str(link.get("#text", "")).lower()

        # Filter links if hash found in component
        for hash_ in component.hashes:
            if str(hash_.alg) == "SHA-256":
                links = jmespath.search(
                    expression=f"""[? ends_with("@href", '#sha256={hash_.content}')]""",
                    data=links,
                )

                break

        return links

    def _map_download_urls(self, component: Component, links: list[dict[str, str]]) -> dict[str, str]:
        pkg_version = component.version or hoppr.utils.get_package_url(component.purl).version

        # Get source links
        source_links = [
            link for link in links
            if (text := link.get("#text", "")).endswith(".tar.gz")
            and f"-{pkg_version}." in text
        ]  # fmt: skip

        # Get binary_links
        binary_links = [
            link for link in links
            if (text := link.get("#text", "")).endswith(".whl")
            and f"-{pkg_version}-" in text
        ]  # fmt: skip

        return {link["#text"].lower(): link["@href"] for link in [*binary_links, *source_links]}

    def _get_file_paths(self, target_dir: Path) -> Iterator[Path]:
        yield from (file_path for file_path in target_dir.iterdir() if file_path.is_file())

    def _update_hashes(self, comp: Component, package_path: Path):
        """Update the hashes of a component with hashes generated from its package file.

        Args:
            comp: Component object whose hashes will be updated.
            package_path: Path of the downloaded package file for the provided component.

        Raises:
            HopprPluginError: If an existing hash doesn't match.
        """
        generated_hashes = list(self._get_artifact_hashes(package_path))

        try:
            comp.update_hashes(generated_hashes)
        except ValueError as ex:
            self.get_logger().debug(
                "Computed hashes: [%s], SBOM hashes: [%s]",
                ", ".join(hash.json() for hash in generated_hashes),
                ", ".join(hash.json() for hash in comp.hashes),
                indent_level=2,
            )
            raise HopprPluginError(f"Hash for {comp.name} does not match expected hash.") from ex

    def _validate_hashes(self, comp: Component, hash_content: str):
        """Validates the component hashes against the hashes in the package metadata provided by the PyPI server.

        Args:
            comp: Component object whose hashes will be validated.
            filename: Name of the package file. Used for querying the metadata.
            hash_content: Content of hash to validate.
            creds: Credentials used for HTTP requests.

        Raises:
            HopprPluginError: If hash validation fails.
        """
        if not comp.contains_hashes([Hash(alg=cdx.HashAlg.SHA_256, content=hash_content)]):
            self.get_logger().debug(
                "Computed hashes: [%s], expected hashes: [%s]",
                ", ".join(hash.json() for hash in comp.hashes),
                hash_content,
                indent_level=2,
            )

            raise HopprPluginError(f"Hash validation failed for {comp.name}.")

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        """Copy a component to the local collection directory structure."""
        # sourcery skip: low-code-quality
        if importlib.util.find_spec(name="pip") is None:
            return Result.fail(message="The pip package was not found. Please install and try again.")

        purl = hoppr.utils.get_package_url(comp.purl)

        links = self._get_download_urls(comp, repo_url, creds)
        file_link_map = self._map_download_urls(comp, links)

        target_dir = self.directory_for(purl.type, repo_url, subdir=f"{purl.name}_{purl.version}")
        self.get_logger().info(msg=f"Target directory: {target_dir}", indent_level=2)

        simple_index_url = RepositoryUrl(url=repo_url)
        if not re.match(pattern="^.*simple/?$", string=str(simple_index_url)):
            simple_index_url /= "simple"

        try:
            success_message = self._download_package(purl, str(simple_index_url), target_dir, creds)
            downloaded_files = self._get_file_paths(target_dir)

            for file in downloaded_files:
                download_url = file_link_map[file.name.lower()]
                *_, hash_value = download_url.split("#sha256=")
                self._update_hashes(comp, file)
                self._validate_hashes(comp, hash_value)
        except KeyError as ex:
            return Result.fail(f"Download URL not found for filename: {ex}", return_obj=comp)
        except HopprPluginRetriableError as ex:
            return Result.retry(f"Failed to download {purl.name} version {purl.version}. {ex}", return_obj=comp)
        except HopprPluginError as ex:
            return Result.fail(f"Failed to download {purl.name} version {purl.version}. {ex}", return_obj=comp)

        artifact_file, *extra = (file for file in target_dir.iterdir() if file.suffix in [".gz", ".whl"])
        self.set_collection_params(comp, repo_url, artifact_file)

        comp.properties.extend(Property(name=BomProps.COLLECTION_ARTIFACT_FILE, value=file.name) for file in extra)

        return Result.success(success_message, return_obj=comp)

    def _get_success_message(self, success_count: int, result_codes: list[int]) -> str:
        message = ""
        match self.collector_config.type, self.collector_config._type_order, result_codes:
            case ("binary-preferred" | "source-preferred", (first_type, second_type), _):
                message = f"Successfully downloaded {first_type}, {second_type} skipped"
            case ("both-preferred", (first_type, second_type), [1, 0]) if success_count == 1:
                message = f"Only able to download {second_type}"
            case ("both-preferred", (first_type, second_type), [0, 1]) if success_count == 1:
                message = f"Only able to download {first_type}"

        return message

    def _get_retry_exception_message(self, success_count: int) -> str:
        message = ""
        match self.collector_config.type, *self.collector_config._type_order:
            case "both-required", first_type, second_type if success_count == 1:
                message = f"Only able to download {first_type}"
            case "binary-preferred" | "source-preferred" | "both-required" | "both-preferred", first_type, second_type:
                message = f"Unable to download {first_type} or {second_type}."
            case "binary-only" | "source-only", first_type:
                message = f"Unable to collect {first_type}"

        return message

    @staticmethod
    def _pypi_param_for_type(collection_type: str) -> str:
        if collection_type == "source":
            return "--no-binary=:all:"

        return "--only-binary=:all:"



"""JSON file parser to generate mappings of object locations for use in Code Climate issue report."""

from __future__ import annotations

import json
import re

from collections import defaultdict, deque
from dataclasses import dataclass, field
from enum import Enum
from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING, Any, Final

from typing_extensions import Self, overload

from hoppr.exceptions import HopprLoadDataError
from hoppr.models.base import CycloneDXBaseModel
from hoppr.models.validation.code_climate import (
    IssueLocationPositionRange,
    PositionRange,
    PositionReference,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from os import PathLike


__all__ = ["JSONLocationMapper"]


class JSONElementName(str, Enum):
    """Names for elements in a JSON document."""

    # Structural characters
    BEGIN_ARRAY = "["
    BEGIN_OBJECT = "{"
    END_ARRAY = "]"
    END_OBJECT = "}"
    NAME_SEPARATOR = ":"
    VALUE_SEPARATOR = ","

    # Literal names
    TRUE = "true"
    FALSE = "false"
    NULL = "null"

    # Types
    NUMBER = "NUMBER"
    STRING = "STRING"
    ARRAY = "ARRAY"
    OBJECT = "OBJECT"

    # Higher-Level
    MEMBER = "MEMBER"
    VALUE = "VALUE"

    # Whitespace
    HORIZONTAL_WHITESPACE = "H_WS"
    VERTICAL_WHITESPACE = "V_WS"

    # Special
    END_OF_FILE = "EOF"
    ERROR = "ERROR"


@dataclass
class JSONElement:
    """Represents a JSON element."""

    name: JSONElementName
    start_position: int
    end_position: int
    start_line: int
    start_column: int
    end_line: int
    end_column: int
    children: list[JSONElement] = field(default_factory=list)


class JSONLexer:
    """Performs lexical tokenization on JSON strings."""

    TOKEN_REGEX_NAME_PAIRS: Final[list[tuple[re.Pattern, JSONElementName | None]]] = [
        (re.compile(r"([\[\]{},:])"), None),  # Structural tokens
        (re.compile(r"\b(true|false|null)\b"), None),  # Literal name tokens
        (
            re.compile(r'(?:"(?:\\(?:["\\\/bfnrt]|u[a-fA-F0-9]{4})|[^"\\\0-\x1F\x7F]+)*")'),
            JSONElementName.STRING,
        ),
        (
            re.compile(r"((?:-?(?:0|[1-9][0-9]*)(?:\.[0-9]+)?(?:[eE][+-]?[0-9]+)?))"),
            JSONElementName.NUMBER,
        ),
        (
            re.compile(r"([^\S\r\n]+)", flags=re.M),
            JSONElementName.HORIZONTAL_WHITESPACE,
        ),
        (re.compile(r"(\r\n|\r|\n)", flags=re.M), JSONElementName.VERTICAL_WHITESPACE),
    ]

    def __init__(self):
        self.position = 0
        self.line_position = 0
        self.line_number = 1

    def _get_next_token(self, reader: StringIO, line: str) -> JSONElement:
        for regex, token_name in self.TOKEN_REGEX_NAME_PAIRS:
            if matched := regex.match(line):
                end_position = self.position + len(matched[0])

                return JSONElement(
                    token_name or JSONElementName(str(matched[1])),
                    self.position,
                    end_position,
                    start_line=self.line_number,
                    start_column=1 + self.position - self.line_position,
                    end_line=self.line_number,
                    end_column=end_position - self.line_position,
                )

        remaining_text = reader.read()
        eof_position = self.position + len(remaining_text)

        return JSONElement(
            JSONElementName.ERROR,
            self.position + len(remaining_text),
            eof_position,
            start_line=self.line_number,
            start_column=1 + self.position - self.line_position,
            end_line=self.line_number + remaining_text.count("\n") + 1,
            end_column=eof_position - self.position + remaining_text.rfind("\n"),
        )

    def _get_line_tokens(self, reader: StringIO, line: str) -> Iterator[JSONElement]:
        new_token = self._get_next_token(reader, line)

        self.position = new_token.end_position
        reader.seek(self.position)

        if new_token.name == JSONElementName.HORIZONTAL_WHITESPACE:
            return

        if new_token.name == JSONElementName.VERTICAL_WHITESPACE:
            self.line_position = self.position
            self.line_number += 1
            return

        yield new_token

    def get_tokens(self, reader: StringIO) -> Iterator[JSONElement]:
        """Performs lexical tokenization of a JSON document.

        Args:
            reader: Text stream to the JSON document you want to tokenize.

        Yields:
            JSON tokens.
        """
        for line in reader:
            yield from self._get_line_tokens(reader, line)

        yield JSONElement(
            JSONElementName.END_OF_FILE,
            self.position,
            self.position,
            start_line=self.line_number,
            start_column=self.position - self.line_position,
            end_line=self.line_number,
            end_column=self.position - self.line_position,
        )


class JSONParser:
    """Parser for JSON documents."""

    VALUE_ELEMENT_NAMES: Final[list[JSONElementName]] = [
        JSONElementName.OBJECT,
        JSONElementName.ARRAY,
        JSONElementName.NUMBER,
        JSONElementName.STRING,
        JSONElementName.TRUE,
        JSONElementName.FALSE,
        JSONElementName.NULL,
    ]

    ALLOWED_AT_ROOT: Final[list[JSONElementName]] = [
        JSONElementName.BEGIN_ARRAY,
        JSONElementName.BEGIN_OBJECT,
        *VALUE_ELEMENT_NAMES,
    ]

    ERROR_MESSAGE: Final[str] = "Invalid JSON document provided."

    def __init__(self):
        self._parse_stack = deque[JSONElement]()

    def _shift(self, token: JSONElement):
        self._parse_stack.append(token)

    def _reduce(
        self,
        production_name: JSONElementName,
        handle_start_name: JSONElementName | None = None,
    ):
        handle: list[JSONElement] = []
        while element := self._parse_stack.pop():
            if not handle_start_name:
                handle_start_name = element.name

            handle = [element, *handle]

            if element.name == handle_start_name:
                self._parse_stack.append(
                    JSONElement(
                        production_name,
                        handle[0].start_position,
                        handle[-1].end_position,
                        handle[0].start_line,
                        handle[0].start_column,
                        handle[-1].end_line,
                        handle[-1].end_column,
                        children=handle,
                    )
                )

                return

    def _reduce_if_member(self):
        if len(self._parse_stack) > 1 and self._parse_stack[-2].name == JSONElementName.NAME_SEPARATOR:
            self._reduce(JSONElementName.MEMBER, JSONElementName.STRING)

    def _handle_token(self, token: JSONElement) -> JSONElement | None:
        parsed: JSONElement | None = None

        match (self._parse_stack, token):
            # ----- Root Values -----
            case [[], token] if token.name in self.ALLOWED_AT_ROOT:
                self._shift(token)

                if token.name in self.VALUE_ELEMENT_NAMES:
                    self._reduce(JSONElementName.VALUE)

            # ----- Property Values, Array Values, and Members -----
            case [
                [
                    *_,
                    JSONElement(name=JSONElementName.MEMBER),
                    JSONElement(name=JSONElementName.VALUE_SEPARATOR),
                ],
                JSONElement(name=JSONElementName.STRING),
            ]:
                self._shift(token)

            case [
                [
                    *_,
                    JSONElement(name=JSONElementName.NAME_SEPARATOR | JSONElementName.BEGIN_ARRAY),
                ]
                | [
                    *_,
                    JSONElement(name=JSONElementName.VALUE),
                    JSONElement(name=JSONElementName.VALUE_SEPARATOR),
                ],
                token,
            ] if token.name in self.ALLOWED_AT_ROOT:
                self._shift(token)

                if token.name in self.VALUE_ELEMENT_NAMES:
                    self._reduce(JSONElementName.VALUE)
                    self._reduce_if_member()

            # ----- Property Names -----
            case [
                [*_, JSONElement(name=JSONElementName.BEGIN_OBJECT)]
                | [
                    *_,
                    JSONElement(name=JSONElementName.VALUE),
                    JSONElement(name=JSONElementName.VALUE_SEPARATOR),
                ],
                JSONElement(name=JSONElementName.STRING),
            ]:
                self._shift(token)

            # ----- Value Separators -----
            case [
                [
                    *_,
                    JSONElement(),
                    JSONElement(name=JSONElementName.VALUE | JSONElementName.MEMBER),
                ],
                JSONElement(name=JSONElementName.VALUE_SEPARATOR),
            ]:
                self._shift(token)

            # ----- Name Separators -----
            case [
                [*_, JSONElement(name=JSONElementName.STRING)],
                JSONElement(name=JSONElementName.NAME_SEPARATOR),
            ]:
                self._shift(token)

            # ----- Object End -----
            case [
                [
                    *_,
                    JSONElement(name=JSONElementName.BEGIN_OBJECT | JSONElementName.MEMBER),
                ],
                JSONElement(name=JSONElementName.END_OBJECT),
            ]:
                self._shift(token)
                self._reduce(JSONElementName.OBJECT, JSONElementName.BEGIN_OBJECT)
                self._reduce(JSONElementName.VALUE, JSONElementName.OBJECT)
                self._reduce_if_member()

            # ----- Array End -----
            case [
                [*_, JSONElement(name=JSONElementName.BEGIN_ARRAY)]
                | [
                    *_,
                    JSONElement(name=JSONElementName.BEGIN_ARRAY) | JSONElement(name=JSONElementName.VALUE_SEPARATOR),
                    JSONElement(name=JSONElementName.VALUE),
                ],
                JSONElement(name=JSONElementName.END_ARRAY),
            ]:
                self._shift(token)
                self._reduce(JSONElementName.ARRAY, JSONElementName.BEGIN_ARRAY)
                self._reduce(JSONElementName.VALUE, JSONElementName.ARRAY)
                self._reduce_if_member()

            # ----- Accept -----
            case [
                [JSONElement(name=JSONElementName.VALUE)],
                JSONElement(name=JSONElementName.END_OF_FILE),
            ]:
                parsed = self._parse_stack.pop()

            # ----- Error -----
            case _:
                raise ValueError(self.ERROR_MESSAGE) from None

        return parsed

    def parse(self, reader: StringIO) -> JSONElement:
        """Generates an abstract syntax tree of a JSON document.

        Args:
            reader: Text stream to the JSON document you want to parse.

        Raises:
            ValueError: If an invalid JSON document is provided.

        Returns:
            An abstract syntax tree with the positions of each element.
        """
        for token in JSONLexer().get_tokens(reader):
            if parsed := self._handle_token(token):
                return parsed

        # Unreachable
        raise ValueError(self.ERROR_MESSAGE)  # pragma: no cover


class JSONLocationMapper:
    """Represents mappings of objects to their locations in a JSON document."""

    def __init__(
        self,
        location_map: defaultdict[str, list[IssueLocationPositionRange]] | None = None,
    ):
        self._location_map = location_map or defaultdict[str, list[IssueLocationPositionRange]](list)

    def __delitem__(self, key: str) -> None:
        del self._location_map[json.dumps(json.loads(key), ensure_ascii=False)]  # pragma: no cover

    def __getitem__(self, key: str) -> list[IssueLocationPositionRange]:
        return self._location_map[json.dumps(json.loads(key), ensure_ascii=False)]

    def _populate_location_map(self, syntax_tree_node: JSONElement, source_path: Path, reader: StringIO):
        if syntax_tree_node.name in {JSONElementName.VALUE, JSONElementName.MEMBER}:
            reader.seek(syntax_tree_node.start_position)
            token_length = syntax_tree_node.end_position - syntax_tree_node.start_position
            location_map_key = reader.read(token_length)

            if syntax_tree_node.name == JSONElementName.MEMBER:
                location_map_key = f"{{{location_map_key}}}"

            location_map_key = json.dumps(json.loads(location_map_key), ensure_ascii=False)

            self._location_map[location_map_key].append(
                IssueLocationPositionRange(
                    path=source_path,
                    positions=PositionRange(
                        begin=PositionReference(
                            line=syntax_tree_node.start_line,
                            column=syntax_tree_node.start_column,
                        ),
                        end=PositionReference(
                            line=syntax_tree_node.end_line,
                            column=syntax_tree_node.end_column,
                        ),
                    ),
                )
            )

        for node in syntax_tree_node.children:
            self._populate_location_map(node, source_path, reader)

    @overload
    def find(self, obj: str) -> list[IssueLocationPositionRange]: ...

    @overload
    def find(self, obj: dict[str, Any]) -> list[IssueLocationPositionRange]: ...

    @overload
    def find(self, obj: list[Any]) -> list[IssueLocationPositionRange]: ...

    @overload
    def find(self, obj: CycloneDXBaseModel) -> list[IssueLocationPositionRange]: ...

    def find(self, obj: str | dict | list | CycloneDXBaseModel | object) -> list[IssueLocationPositionRange]:
        """Get file position information for an object in the parser's file.

        Args:
            obj: Object to search for.

        Raises:
            HopprLoadDataError: Attempted to load invalid JSON string

        Returns:
            A list of `IssueLocationPositionRange` containing start and end locations of the referenced object.
        """
        try:
            match obj:
                case dict() | list():
                    key = json.dumps(obj, ensure_ascii=False)

                case str():
                    key = json.dumps(json.loads(obj), ensure_ascii=False)

                case CycloneDXBaseModel():
                    key = obj.json()

                case _:
                    key = str(obj).replace("'", '"')

            return self._location_map[key]

        except json.JSONDecodeError:
            raise HopprLoadDataError(f"Failed to parse as valid JSON: {obj}") from None

    @classmethod
    def load_file(cls, source_path: PathLike[str] | str) -> Self:
        """Maps JSON values and members to their line and column numbers.

        Args:
            source_path: Path to the source file containing JSON document to be mapped.

        Returns:
            Mapping of parsed JSON objects to their location in the file
        """
        source_path = Path(source_path)
        location_mapper = cls()

        reader = StringIO(initial_value=source_path.read_bytes().decode(encoding="utf-8"))

        try:
            syntax_tree = JSONParser().parse(reader)
            location_mapper._populate_location_map(syntax_tree, source_path, reader)
        finally:
            reader.close()

        return location_mapper



"""CycloneDX data models."""

from __future__ import annotations

import uuid

from collections.abc import Callable, MutableMapping
from datetime import datetime, timezone
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, ClassVar, Literal, TypeAlias

import hoppr_cyclonedx_models.cyclonedx_1_5 as cdx

from pydantic import Extra, Field, root_validator, validator
from rapidfuzz import fuzz
from requests import HTTPError

import hoppr.net
import hoppr.utils

from hoppr.exceptions import HopprLoadDataError
from hoppr.models.affect import Affect
from hoppr.models.annotations import Annotations
from hoppr.models.base import CycloneDXBaseModel, UniqueIDMap
from hoppr.models.licenses import LicenseChoice, LicenseExpressionItem, LicenseMultipleItem, NamedLicense
from hoppr.models.types import LocalFile, OciFile, UrlFile

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from packageurl import PackageURL


DictStrAny: TypeAlias = dict[str, Any]
UpdateRefsCallback: TypeAlias = Callable[[str], None]

# Type aliases for CycloneDX enums
ComponentType: TypeAlias = cdx.Type
PatchType: TypeAlias = cdx.Type1
IssueType: TypeAlias = cdx.Type2
ExternalReferenceType: TypeAlias = cdx.Type3
LearningType: TypeAlias = cdx.Type4
SubjectMatterType: TypeAlias = cdx.Type5
EventType: TypeAlias = cdx.Type6
DataOutputType: TypeAlias = cdx.Type7
__all__ = [
    "Component",
    "ComponentType",
    "DataOutputType",
    "EventType",
    "ExternalReference",
    "ExternalReferenceType",
    "Hash",
    "IssueType",
    "LearningType",
    "Metadata",
    "PatchType",
    "Property",
    "Sbom",
    "SubjectMatterType",
    "Tools",
]

SbomRef = Annotated[LocalFile | OciFile | UrlFile, Field(description="Reference to a local or remote SBOM file")]

FUZZY_MATCH_THRESHOLD = 85


def _extract_components(components: list[Component]) -> list[Component]:
    """Explicitly set scope of flattened components to `exclude`."""
    for component in components:
        component.scope = cdx.Scope.EXCLUDED

    return hoppr.utils.dedup_list(components)


def _extract_sbom_components(external_refs: list[ExternalReference]) -> list[Component]:
    """Extracts `external_refs` of type "bom" and returns the set of their components."""
    components: list[Component] = []

    for ref in _get_bom_refs(external_refs):
        sbom = Sbom.load(_resolve_sbom_source(ref.dict()["url"]))
        sbom.components = _extract_components(sbom.components)
        components.extend(sbom.components)
        external_refs.remove(ref)

    return components


def _flatten_component(component: Component) -> list[Component]:
    """Helper function to flatten a component's subcomponents into a set."""
    flattened = []

    for subcomponent in component.components or []:
        # Ensure validator is run to set `bom_ref`
        subcomponent = Component(**subcomponent.dict())
        subcomponent.scope = cdx.Scope.EXCLUDED

        # Flatten nested components into top level components list
        flattened.append(subcomponent)

    component.components.clear()
    return flattened


def _get_bom_refs(external_refs: list[ExternalReference]) -> Iterator[ExternalReference]:
    """Get `externalReferences` of type "bom"."""
    yield from (ref.copy(deep=True) for ref in (external_refs or []) if ref.type == "bom")


def _make_validator(function: Callable, field_name: str, pre: bool = False) -> classmethod[Any, Any, Any]:
    """Wrap a callable with the pydantic validator wrapper."""
    return validator(field_name, allow_reuse=True, always=True, pre=pre)(function)


def _resolve_sbom_source(source: str) -> str | Path | DictStrAny:
    """Resolves an SBOM source as a file path, URL or `dict`."""
    return Path(source.removeprefix("file://")).resolve() if source.startswith("file://") else source


def _set_component_bom_ref(values: DictStrAny) -> DictStrAny:
    """Set `bom-ref` identifier for a component if not set."""
    component = Component.parse_obj(values)

    if not any([component.purl, component.bom_ref, all([component.name, component.version])]):
        raise ValueError(
            "Either 'bom-ref' or 'purl' must be defined, or 'name' and 'version' must be defined on a component"
        )

    if component.purl:
        # Decode any unicode escape sequences, e.g. "\u0026" -> "&"
        component.purl = component.purl.encode("utf-8").decode("utf-8")

        # If component has no `version` set, try to parse version from `purl` field
        purl_version = hoppr.utils.get_package_url(component.purl).version
        component.version = component.version or purl_version

    bom_ref = str(
        component.bom_ref or component.purl or f"{'@'.join(filter(None, [component.name, component.version]))}"
    )

    component.bom_ref = bom_ref.encode("utf-8").decode("utf-8")

    return component.dict(exclude_none=True, exclude_unset=True)


def _validate_components_pre(cls: CycloneDXBaseModel, values: list[DictStrAny]) -> list[DictStrAny]:
    """Validator to set `bom-ref` identifier for each component if not set.

    This validation is only performed on the `components` field for the `Sbom` and `Component` models.
    """
    validated: list[DictStrAny] = [_set_component_bom_ref(component_dict) for component_dict in values]

    return validated


def _validate_components(cls: CycloneDXBaseModel, components: list[Component]) -> list[Component]:
    """Validator to optionally flatten `components` list."""
    if not cls.flatten:
        return components

    flattened = list(components)

    for component in components:
        flattened.extend(_flatten_component(component))

    return hoppr.utils.dedup_list(flattened)


def _validate_external_refs(
    cls: CycloneDXBaseModel, external_refs: list[ExternalReference], values: DictStrAny
) -> list[ExternalReference]:
    """Validator to optionally resolve `externalReferences`."""
    external_refs = [ExternalReference.create(ref) for ref in external_refs or []]

    if cls.deep_merge:
        external_ref_components = _extract_sbom_components(external_refs)
        values["components"] = hoppr.utils.dedup_list([*values.get("components", []), *external_ref_components])

    return external_refs


def _validate_licenses(cls: CycloneDXBaseModel, licenses: LicenseChoice) -> list[LicenseMultipleItem]:
    """Validator to resolve expression field."""
    license_list: list[LicenseMultipleItem] = [
        LicenseMultipleItem(license=NamedLicense(bom_ref=license_.bom_ref, name=license_.expression))
        if isinstance(license_, LicenseExpressionItem)
        else license_
        for license_ in licenses or []
    ]

    return hoppr.utils.dedup_list(license_list)


class ExternalReference(CycloneDXBaseModel, cdx.ExternalReference, extra=Extra.allow):
    """ExternalReference data model derived from CycloneDXBaseModel."""

    type: ExternalReferenceType

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}


class Hash(CycloneDXBaseModel, cdx.Hash, extra=Extra.allow):
    """Hash data model derived from CycloneDXBaseModel."""

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}


class Property(CycloneDXBaseModel, cdx.Property, extra=Extra.allow):
    """Property data model derived from CycloneDXBaseModel."""

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}


class Component(CycloneDXBaseModel, cdx.Component, extra=Extra.allow):
    """Component data model derived from CycloneDXBaseModel."""

    type: ComponentType
    bom_ref: Annotated[str, Field(alias="bom-ref")] = None  # type: ignore[assignment]
    components: list[Component] = []  # type: ignore[assignment]
    externalReferences: list[ExternalReference] = []  # type: ignore[assignment]
    hashes: list[Hash] = []  # type: ignore[assignment]
    licenses: LicenseChoice = []
    properties: list[Property] = []  # type: ignore[assignment]

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}

    validate_components_pre: classmethod = _make_validator(_validate_components_pre, "components", pre=True)
    validate_components: classmethod = _make_validator(_validate_components, "components")
    validate_external_refs: classmethod = _make_validator(_validate_external_refs, "externalReferences")

    reference_updater_map: ClassVar[MutableMapping[Component, UpdateRefsCallback]] = {}

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Component):
            return False

        # Remove `purl` and `bom_ref` from field comparisons. Field checks are only
        # performed if these fields have already been checked and found to be not equal
        field_names = (self.__fields_set__ | other.__fields_set__).difference({"bom_ref", "purl"})

        return (
            self._purl_check(other)
            or self._bom_ref_check(other)
            or all(getattr(self, name, None) == getattr(other, name, None) for name in field_names)
        )

    def _bom_ref_check(self, other: Component) -> bool:
        if not (self.bom_ref and other.bom_ref):
            return False

        return self.bom_ref == other.bom_ref and self._hash_match(other.hashes)

    def _hash_match(self, other_hashes: list[Hash]) -> bool:
        """Compare this component's hashes against a specified list of hashes.

        Args:
            other_hashes: Hash list to compare against this component's hashes.

        Returns:
            `True` if the content of each component hash matches the content of each hash in
            `other_hashes` with a matching algorithm; otherwise, `False`.
        """
        content_match = True

        # Compare hashes from both components that have matching algorithm
        hash_alg_matches = (
            (self_hash, other_hash)
            for self_hash in self.hashes
            for other_hash in other_hashes
            if self_hash.alg == other_hash.alg
        )

        for self_hash, other_hash in hash_alg_matches:
            content_match = content_match and self_hash.content == other_hash.content

        return content_match

    def _purl_check(self, other: Component) -> bool:
        if not (self.purl and other.purl):
            return False

        self_purl = hoppr.utils.get_package_url(self.purl)
        other_purl = hoppr.utils.get_package_url(other.purl)

        qual_keys = hoppr.utils.dedup_list(self_purl.qualifiers | other_purl.qualifiers)

        return all([
            self_purl.name == other_purl.name,
            self_purl.type == other_purl.type,
            self_purl.namespace == other_purl.namespace,
            str(self_purl.version).removeprefix("v") == str(other_purl.version).removeprefix("v"),
            self_purl.subpath == other_purl.subpath,
            *[self._qualifier_match(key, self_purl, other_purl) for key in qual_keys],
        ]) and self._hash_match(other.hashes)

    def _qualifier_match(self, key: str, self_purl: PackageURL, other_purl: PackageURL) -> bool:
        # Compare only if both purls have a value for specified qualifier
        if key not in (self_qual := self_purl.qualifiers) or key not in (other_qual := other_purl.qualifiers):
            return True

        return fuzz.ratio(self_qual.get(key, ""), other_qual.get(key, "")) > FUZZY_MATCH_THRESHOLD

    def contains_hashes(self, other_hashes: list[Hash]) -> bool:
        """Verify that the provided hashes are in this component's hashes.

        Args:
            other_hashes: Hash list used to validate this component's hashes.

        Returns:
            `True` if the provided hashes match all component hashes; otherwise, `False`.
        """
        return all(other_hash in self.hashes for other_hash in other_hashes)

    def update_hashes(self, hashes: Iterable[Hash]):
        """Updates the hashes for the component.

        Hashes using an algorithm matching an existing hash in this component will be compared against the current hash
        for that algorithm. If they don't match, a `ValueError` will be raised.

        Args:
            hashes: Hashes for updating the component hashes.

        Raises:
            ValueError: If the provided hashes have an existing algorithm that doesn't match the corresponding hash in
            the component.
        """
        current_hash_algs = [hash_.alg for hash_ in self.hashes]
        for hash_ in hashes:
            if hash_.alg not in current_hash_algs:
                self.hashes.append(hash_)
            elif hash_ not in self.hashes:
                raise ValueError("Mismatched hash for existing algorithm.")

    def merge(self, other: CycloneDXBaseModel) -> None:
        """Merge Component instance into self.

        Args:
            other: Component object to merge
        """
        if not isinstance(other, Component):
            return

        if other.bom_ref and self.bom_ref != other.bom_ref and other in self.reference_updater_map:
            self.reference_updater_map[other](self.bom_ref)

        super().merge(other)


Component.update_forward_refs()


class Tools(CycloneDXBaseModel, cdx.ToolModel, extra=Extra.allow):
    """Tools data model derived from CycloneDXBaseModel."""

    components: list[Component] = Field(default=[])  # type: ignore[assignment]

    @classmethod
    def _convert_tool_to_component(cls, tool: cdx.Tool) -> Component:
        return Component(
            type=ComponentType.APPLICATION,
            name=str(tool.name),
            version=tool.version,
            hashes=[Hash.create(hash) for hash in tool.hashes or []],
            externalReferences=[ExternalReference.create(ref) for ref in tool.externalReferences or []],
            scope=cdx.Scope.EXCLUDED,
        )

    @root_validator(allow_reuse=True, pre=True)
    @classmethod
    def validate_tools(cls, values: list[DictStrAny] | DictStrAny) -> DictStrAny:
        """Validator to convert deprecated list of `Tool` objects to a `Tools` object."""
        if isinstance(values, list):
            tool_model = cdx.ToolModel()
            tool_model.components = []

            tools = [cdx.Tool.parse_obj(tool) for tool in values]
            tool_model.components.extend(cls._convert_tool_to_component(tool) for tool in tools)

            values = tool_model.dict(exclude_none=True, exclude_unset=True)

        return values

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}


class Metadata(CycloneDXBaseModel, cdx.Metadata, extra=Extra.allow):
    """Metadata data model derived from CycloneDXBaseModel."""

    tools: Tools | None = None
    licenses: LicenseChoice = None

    validate_licenses: classmethod = _make_validator(_validate_licenses, "licenses")

    @root_validator(allow_reuse=True, pre=True)
    @classmethod
    def validate_metadata(cls, values: DictStrAny | None) -> DictStrAny:
        """Validator to populate and normalize `metadata` field."""
        values = values or {}

        if isinstance(tool_list := values.get("tools"), list):
            values["tools"] = Tools.validate_tools(tool_list)

        metadata = cdx.Metadata.parse_obj(values)
        metadata.timestamp = datetime.now(timezone.utc)
        metadata.tools = Tools.parse_obj(values.get("tools") or {})

        # Generate `metadata` field containing Hoppr tool component
        metadata.tools.components = hoppr.utils.dedup_list([
            Component(
                type=cdx.Type("application"),
                name="hoppr",
                version=hoppr.__version__,
                bom_ref=f"pkg:pypi/hoppr@{hoppr.__version__}",
                purl=f"pkg:pypi/hoppr@{hoppr.__version__}",
                scope=cdx.Scope.EXCLUDED,
            ),
            *metadata.tools.components,
        ])

        values = metadata.dict(exclude_none=True, exclude_unset=True)
        return values

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}


class Vulnerability(CycloneDXBaseModel, cdx.Vulnerability, extra=Extra.allow):
    """Vulnerability data model derived from CycloneDXBaseModel."""

    tools: Tools | None = None  # type: ignore[assignment]
    affects: Annotated[
        list[Affect],
        Field(description="The components or services that are affected by the vulnerability.", title="Affects"),
    ] = []  # type: ignore[assignment]

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}


class Sbom(CycloneDXBaseModel, cdx.CyclonedxSoftwareBillOfMaterialsStandard, extra=Extra.allow):
    """Sbom data model derived from CycloneDXBaseModel."""

    specVersion: str = "1.5"
    version: int = 1
    metadata: Metadata | None = None
    components: list[Component] = []  # type: ignore[assignment]
    externalReferences: list[ExternalReference] = []  # type: ignore[assignment]
    vulnerabilities: list[Vulnerability] = []  # type: ignore[assignment]
    annotations: list[Annotations] = []  # type: ignore[assignment]

    # Attributes not included in schema
    loaded_sboms: ClassVar[MutableMapping[SbomRef, "Sbom"]] = {}
    unique_id_map: ClassVar[UniqueIDMap] = {}

    validate_components_pre: classmethod = _make_validator(_validate_components_pre, "components", pre=True)
    validate_components: classmethod = _make_validator(_validate_components, "components")
    validate_external_refs: classmethod = _make_validator(_validate_external_refs, "externalReferences")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Sbom):
            return False

        return True if self.serialNumber == other.serialNumber else super().__eq__(other)

    def __hash__(self) -> int:
        return hash(self.serialNumber)

    @root_validator(allow_reuse=True, pre=True)
    @classmethod
    def validate_sbom(cls, values: DictStrAny) -> DictStrAny:
        """Validator to standardize fields."""
        values["$schema"] = "http://cyclonedx.org/schema/bom-1.5.schema.json"
        values["specVersion"] = "1.5"
        values["serialNumber"] = values.get("serialNumber", None) or uuid.uuid4().urn

        return values

    @classmethod
    def find_ref(cls, ref_type: Literal["local", "oci", "url"], location: str | Path) -> Sbom | None:
        """Look up SBOM object by reference.

        Args:
            ref_type (Literal["local", "oci", "url"]): Type of SBOM reference
            location (str | Path): Location of SBOM reference

        Returns:
            Sbom | None: SBOM object if found, otherwise None
        """
        match ref_type:
            case "local":
                return cls.loaded_sboms.get(LocalFile(local=Path(location)), None)
            case "oci":
                return cls.loaded_sboms.get(OciFile(oci=str(location)), None)
            case "url":
                return cls.loaded_sboms.get(UrlFile(url=str(location)), None)
            case _:
                return None

    @classmethod
    def load(cls, source: str | Path | DictStrAny) -> Sbom:
        """Load SBOM from local file, URL, or dict."""
        match source:
            case dict():
                sbom = cls(**source)
            case Path():
                # Convert source to relative path if in current working directory subpath
                source = source.resolve()
                source = source.relative_to(Path.cwd()) if source.is_relative_to(Path.cwd()) else source

                sbom = cls.parse_file(source)
                cls.loaded_sboms[LocalFile(local=source)] = sbom
            case str():
                try:
                    sbom_dict = hoppr.net.load_url(source)
                    if not isinstance(sbom_dict, dict):
                        raise TypeError("URL SBOM was not loaded as dictionary")

                    sbom = cls.parse_obj(sbom_dict)
                    url_ref = UrlFile(url=source)
                    cls.loaded_sboms[url_ref] = sbom
                except (HopprLoadDataError, HTTPError) as ex:
                    raise HopprLoadDataError from ex

        return sbom

    @staticmethod
    def _append_dependency_update_functions(
        update_functions: list[UpdateRefsCallback], dependency: cdx.Dependency, bom_ref: str
    ):
        if dependency.ref == bom_ref:

            def update_dependency_ref(new_ref: str):
                dependency.ref = new_ref

            update_functions.append(update_dependency_ref)

        bom_ref_link = cdx.RefLinkType(bom_ref)
        subdependencies: list[cdx.RefLinkType] = dependency.dependsOn or []
        if bom_ref_link in subdependencies:

            def update_subdependency(new_ref: str):
                subdependencies.remove(bom_ref_link)
                subdependencies.append(cdx.RefLinkType(new_ref))

            update_functions.append(update_subdependency)

    @staticmethod
    def _append_vulnerability_update_function(
        update_functions: list[UpdateRefsCallback], vulnerability: Vulnerability, bom_ref: str
    ):
        def update_affected_ref(new_ref: str, affected: Affect):
            affected.ref = new_ref

        for affected in filter(lambda af: af.ref == bom_ref, vulnerability.affects):
            update_functions.append(partial(update_affected_ref, affected=affected))

    def _get_update_refs_callback(self, bom_ref: str) -> UpdateRefsCallback:
        update_functions: list[UpdateRefsCallback] = []

        def update_refs(new_ref: str):
            for update in update_functions:
                update(new_ref)

        for dependency in self.dependencies or []:
            self._append_dependency_update_functions(update_functions, dependency, bom_ref)

        for vulnerability in self.vulnerabilities:
            self._append_vulnerability_update_function(update_functions, vulnerability, bom_ref)

        return update_refs

    def _set_component_bom_ref_update_callbacks(self):
        for component in self.components:
            Component.reference_updater_map[component] = self._get_update_refs_callback(component.bom_ref)

    def merge(self, other: CycloneDXBaseModel) -> None:
        """Merge Sbom instance into self.

        Args:
            other: Sbom object to merge
        """
        if type(other) is Sbom:
            other._set_component_bom_ref_update_callbacks()

        super().merge(other)



"""Models to perform SBOM validation checks."""

from __future__ import annotations

import json
import logging
import os
import re

from datetime import date, timedelta
from typing import TYPE_CHECKING, Any, TypeAlias

import hoppr_cyclonedx_models.cyclonedx_1_5 as cdx
import jmespath

from pydantic import ValidationError, root_validator, validator

from hoppr.exceptions import HopprValidationError
from hoppr.models.licenses import LicenseExpressionItem, LicenseMultipleItem
from hoppr.models.validation.base import BaseValidator
from hoppr.models.validation.checks import ValidateConfig
from hoppr.models.validation.code_climate import Issue, IssueList, IssueSeverity
from hoppr.models.validation.exclude import JMESPathOptions
from hoppr.models.validation.json_mapper import JSONLocationMapper

if TYPE_CHECKING:
    from pathlib import Path

    from pydantic.fields import ModelField

    DictStrAny: TypeAlias = dict[str, Any]

_CONFIG: ValidateConfig


def _field_validator(*field_names: str, check_name: str) -> classmethod[Any, Any, Any]:
    """Wrap a callable with the pydantic `validator` or `root_validator` decorator."""

    def _validator(obj: object | None, field: ModelField, values: DictStrAny) -> object:
        # Remove current field from `field_names`
        fields = list(field_names)
        fields.remove(field.name)

        # The `values` variable is a dictionary of all fields previously validated by pydantic.
        # This check will be `False` if this is the last unvalidated field in `field_names`.
        if any(name not in values for name in fields):
            return obj

        # Current field plus previously validated fields
        field_values = [obj, *[values.get(name) for name in fields]]

        if not any(field_values):
            missing_fields = ", ".join(f"'{name}'" for name in field_names)
            msg = f"Missing field{'s: one of' if len(field_names) > 1 else ':'} {missing_fields}"

            BaseValidator.notify(msg=msg, log_level=_get_log_level(check_name), indent_level=2)

            raise HopprValidationError(msg, check_name=check_name)

        return obj

    # Assign validator method a unique name
    _validator.__name__ = f"_validate_{'_'.join(field_names)}"

    return validator(*field_names, allow_reuse=True, always=True, pre=True)(_validator)


def _get_check_config(check_name: str) -> str:
    match check_name:
        case _ if matched := re.match(pattern=r"components\.licenses\.(.*)", string=check_name):
            check_config = _CONFIG.component_checks.licenses.dict(by_alias=True)[matched[1]]

        case _ if matched := re.match(pattern=r"components\.(.*)", string=check_name):
            check_config = _CONFIG.component_checks.dict(by_alias=True)[matched[1]]

        case _ if matched := re.match(pattern=r"metadata\.licenses\.(.*)", string=check_name):
            check_config = _CONFIG.metadata_checks.licenses.dict(by_alias=True)[matched[1]]

        case _ if matched := re.match(pattern=r"metadata\.(.*)", string=check_name):
            check_config = _CONFIG.metadata_checks.dict(by_alias=True)[matched[1]]

        case _ if matched := re.match(pattern=r"sbom\.(.*)", string=check_name):
            check_config = _CONFIG.sbom_checks.dict(by_alias=True)[matched[1]]

        case _:
            check_config = "ignore"

    return check_config


def _get_file_issue_list(exc: ValidationError, sbom_file: Path) -> IssueList:
    issue_list = IssueList()
    location_mapper = _get_location_map(sbom_file)

    for error in exc.errors():
        ctx = error.get("ctx", {})
        check_name = ctx.get("check_name", "")

        # Prepend `components` or `metadata` to the check name to
        # identify which type of license check it pertains to
        if check_name.startswith("licenses."):
            loc_idx = error["loc"].index("licenses")
            loc = tuple(reversed(error["loc"][:loc_idx]))
            prefix = next((name for name in loc if name in {"components", "metadata"}), None)
            check_name = f"{f'{prefix}.' if prefix else ''}{check_name}"

        if not (locations := location_mapper.find(_get_file_location_from_error(error["loc"], sbom_file))):
            continue

        location, *_ = locations

        if (severity := _get_severity(check_name)) == IssueSeverity.INFO:
            continue

        issue_list.append(
            Issue(
                check_name=check_name,
                description=error.get("msg", ""),
                location=location,
                severity=severity,
            )
        )

    return issue_list


def _get_file_location_from_error(error_loc: tuple[int | str, ...], sbom_file: Path) -> DictStrAny:
    """Get the location of the failing object in the specified SBOM file.

    For example:
        error_loc = ("components", 2, "licenses", 1, "__root__")

        corresponds to

        file_loc["components"][2]["licenses"][1]

    Args:
        error_loc: A pydantic ErrorDict["loc"] describing the JSON path to the failing object.
        sbom_file: Path to SBOM file with a validation issue.

    Returns:
        Dict representing the the object in which a validation issue was found.
    """
    file_loc = json.loads(sbom_file.read_bytes().decode(encoding="utf-8"))

    for sub_loc in error_loc:
        match sub_loc:
            case str():
                file_loc = file_loc.get(sub_loc, file_loc)

                if isinstance(file_loc, str):
                    return {sub_loc: file_loc}

            case int():
                file_loc = file_loc[sub_loc]

    return file_loc


def _get_location_map(sbom_file: Path) -> JSONLocationMapper:
    location_mapper = JSONLocationMapper.load_file(sbom_file)
    data = json.loads(sbom_file.read_bytes().decode(encoding="utf-8"))
    searches = _CONFIG.exclude.dict(by_alias=True, exclude_none=True, exclude_unset=True) if _CONFIG.exclude else {}

    # Remove excluded objects from location map
    for exclude_type, search_list in searches.items():
        search = "||".join(expr.removeprefix("jmespath:") for expr in search_list)
        search = f"{exclude_type}[*]|{search.replace(']||[?', '||')}"

        for result in jmespath.search(
            expression=search.removeprefix("jmespath:"),
            data=data,
            options=JMESPathOptions(),
        ) or []:  # fmt: skip
            del location_mapper[json.dumps(result, ensure_ascii=False)]

    return location_mapper


def _get_log_level(check_name: str) -> int:
    log_mapping = {
        "ignore": logging.INFO,
        "warn": logging.WARN,
        "error": logging.ERROR,
    }

    return log_mapping.get(_get_check_config(check_name), logging.INFO)


def _get_severity(check_name: str) -> IssueSeverity:
    severity_mapping = {
        "ignore": IssueSeverity.INFO,
        "warn": IssueSeverity.MINOR,
        "error": IssueSeverity.MAJOR,
    }

    return severity_mapping.get(_get_check_config(check_name), IssueSeverity.INFO)


class LicenseValidator(BaseValidator):
    """Model to perform validation checks on licenses."""

    __root__: LicenseMultipleItem | LicenseExpressionItem

    @root_validator
    @classmethod
    def _validate_license_expiration(cls, values: DictStrAny) -> DictStrAny:
        license_ = values["__root__"]

        expiration_days = float(os.getenv("HOPPR_EXPIRATION_DAYS", "30"))

        if (
            isinstance(license_, LicenseMultipleItem)
            and license_.license.licensing
            and license_.license.licensing.expiration
            and license_.license.licensing.expiration.date() < date.today() + timedelta(days=expiration_days)
        ):
            raise HopprValidationError(
                f"License expired or expiring within {int(expiration_days)} days",
                check_name="licenses.expiration",
            )

        return values

    @root_validator
    @classmethod
    def _validate_last_renewal(cls, values: DictStrAny) -> DictStrAny:
        license_ = values["__root__"]

        if (
            isinstance(license_, LicenseMultipleItem)
            and license_.license.licensing
            and license_.license.licensing.lastRenewal
        ):
            return values

        raise HopprValidationError("Missing or invalid lastRenewal", check_name="licenses.last-renewal")

    @root_validator
    @classmethod
    def _validate_license_types(cls, values: DictStrAny) -> DictStrAny:
        license_ = values["__root__"]

        if (
            isinstance(license_, LicenseMultipleItem)
            and license_.license.licensing
            and license_.license.licensing.licenseTypes
        ):
            return values

        raise HopprValidationError("Missing or invalid licenseTypes", check_name="licenses.license-types")

    @root_validator
    @classmethod
    def _validate_purchase_order(cls, values: DictStrAny) -> DictStrAny:
        license_ = values["__root__"]

        if (
            isinstance(license_, LicenseMultipleItem)
            and license_.license.licensing
            and license_.license.licensing.purchaseOrder
        ):
            return values

        raise HopprValidationError("Missing or invalid purchaseOrder", check_name="licenses.purchase-order")

    @root_validator
    @classmethod
    def _validate_license_name_or_id(cls, values: DictStrAny) -> DictStrAny:
        license_ = values["__root__"]

        if isinstance(license_, LicenseMultipleItem) and (license_.license.name or license_.license.id):
            return values

        raise HopprValidationError("Missing license name or id", check_name="licenses.name-or-id")


LicenseValidator.update_forward_refs()


class ComponentValidator(BaseValidator, cdx.Component):
    """Model to perform validation checks on components."""

    licenses: list[LicenseValidator] | None = None
    components: list[ComponentValidator] | None = None  # type: ignore[assignment]

    @root_validator(allow_reuse=True, pre=True)
    @classmethod
    def _validate(cls, values: DictStrAny) -> DictStrAny:
        component_id = "@".join(filter(None, [values.get("name"), values.get("version")]))

        BaseValidator.notify(
            f"Validating component: {component_id}",
            log_level=logging.INFO,
            indent_level=1,
        )

        for license_ in values.get("licenses") or []:
            try:
                LicenseValidator.validate(license_)
            except ValidationError as ex:
                license_id = (
                    license_.get("expression")
                    if "expression" in license_
                    else str(license_.get("license").get("id") or license_.get("license").get("name"))
                )

                BaseValidator.notify(
                    f"Issues found for license {license_id}:",
                    log_level=logging.INFO,
                    indent_level=2,
                )

                for error in ex.errors():
                    ctx = error.get("ctx", {})
                    check_name = ctx.get("check_name", "")

                    BaseValidator.notify(
                        msg=str(error.get("msg")),
                        log_level=_get_log_level(f"components.{check_name}"),
                        indent_level=3,
                    )

        return values

    _validate_licenses_field = _field_validator("licenses", check_name="components.licenses.licenses-field")
    _validate_supplier = _field_validator("supplier", check_name="components.supplier-field")
    _validate_name = _field_validator("name", check_name="components.name-field")
    _validate_version = _field_validator("version", check_name="components.version-field")
    _validate_unique_id = _field_validator("cpe", "purl", "swid", check_name="components.unique-id")


ComponentValidator.update_forward_refs()


class MetadataValidator(BaseValidator, cdx.Metadata):
    """Model to perform validation checks on SBOM metadata."""

    timestamp: str | None = None  # type: ignore[assignment]
    tools: list | dict | None = None  # type: ignore[assignment]
    component: ComponentValidator | None = None
    licenses: list[LicenseValidator] | None = None

    @root_validator(allow_reuse=True, pre=True)
    @classmethod
    def _validate(cls, values: DictStrAny) -> DictStrAny:
        BaseValidator.notify("Validating Metadata", indent_level=1)
        for license_ in values.get("licenses") or []:
            try:
                LicenseValidator.validate(license_)
            except ValidationError as ex:
                license_id = (
                    license_.get("expression")
                    if "expression" in license_
                    else str(license_.get("license").get("id") or license_.get("license").get("name"))
                )

                BaseValidator.notify(
                    f"Issues found for license {license_id}:",
                    log_level=logging.INFO,
                    indent_level=2,
                )

                for error in ex.errors():
                    ctx = error.get("ctx", {})
                    check_name = ctx.get("check_name", "")

                    BaseValidator.notify(
                        msg=str(error.get("msg")),
                        log_level=_get_log_level(f"components.{check_name}"),
                        indent_level=3,
                    )

        return values

    _validate_timestamp = _field_validator("timestamp", check_name="metadata.timestamp")
    _validate_authors = _field_validator("authors", "tools", check_name="metadata.authors")
    _validate_licenses_field = _field_validator("licenses", check_name="metadata.licenses.licenses-field")
    _validate_supplier = _field_validator("supplier", check_name="metadata.supplier-field")


MetadataValidator.update_forward_refs()


class SbomValidator(BaseValidator, cdx.CyclonedxSoftwareBillOfMaterialsStandard):
    """Model to perform validation checks on overall SBOM document."""

    metadata: MetadataValidator | None = None
    components: list[ComponentValidator] | None = None  # type: ignore[assignment]

    _validate_components_field = _field_validator("components", check_name="sbom.components-field")
    _validate_vulnerabilities_field = _field_validator("vulnerabilities", check_name="sbom.vulnerabilities-field")
    _validate_unique_id = _field_validator("serialNumber", check_name="sbom.unique-id")

    @root_validator(pre=True)
    @classmethod
    def _remove_schema_field(cls, values: DictStrAny) -> DictStrAny:
        values.pop("$schema", None)

        return values

    @validator("specVersion", always=True, pre=True)
    @classmethod
    def _validate_spec_version(cls, spec_version: str) -> str:
        match spec_version:
            case "1.2":
                raise HopprValidationError(
                    f"Got specVersion {spec_version}; cannot be validated",
                    check_name="sbom.spec-version",
                )
            case "1.3" | "1.4":
                raise HopprValidationError(
                    f"Got specVersion {spec_version}; should be 1.5",
                    check_name="sbom.spec-version",
                )
            case "1.5":
                return spec_version
            case _:
                raise HopprValidationError(
                    f"Got an invalid specVersion of {spec_version}; should be 1.5",
                    check_name="sbom.spec-version",
                )


SbomValidator.update_forward_refs()


def validate(sbom_file: Path) -> IssueList:
    """Run validation against SBOM(s) with specified configuration.

    Args:
        sbom_file: Path to a file to run validation against.
    """
    # Get previously parsed ValidateConfig. Each call returns
    # the existing config without instantiating a new one.
    global _CONFIG
    _CONFIG = ValidateConfig()

    issue_list = IssueList()

    try:
        SbomValidator.parse_file(sbom_file)
    except ValidationError as exc:
        # Capture errors and create Code Climate issue for each
        issue_list.extend(_get_file_issue_list(exc, sbom_file))

    return issue_list



"""Supporting class for oras bundle plugin."""

from __future__ import annotations

import copy
import json
import os
import platform

from datetime import datetime
from http import cookiejar
from pathlib import Path
from typing import TYPE_CHECKING

import jsonschema
import oras.auth.utils
import oras.container
import oras.defaults
import oras.oci
import oras.provider
import oras.schemas
import oras.utils

from oras import decorator
from oras.container import Container
from oras.decorator import ensure_container
from requests import Response, Session

import hoppr

from hoppr.logger import HopprLogger

if TYPE_CHECKING:
    from oras.container import Container
    from oras.types import container_type


class Registry(oras.provider.Registry):
    """Override the default Oras Registry Class."""

    def __init__(
        self,
        hostname: str | None = None,
        insecure: bool = False,
        tls_verify: bool = True,
        auth_backend: str = "token",
        logger: HopprLogger | None = None,
    ):
        """Create an ORAS client."""
        self.hostname: str | None = hostname
        self.headers: dict = {}
        self.session: Session = Session()
        self.prefix: str = "http" if insecure else "https"
        self._tls_verify = tls_verify
        self.token: str | None = None

        self.logger = logger or HopprLogger(flush_immed=True)

        # Ignore all cookies: some registries try to set one
        # and take it as a sign they are talking to a browser,
        # trying to set further CSRF cookies (Harbor is such a case)
        self.session.cookies.set_policy(cookiejar.DefaultCookiePolicy(allowed_domains=[]))

        # Get custom backend, pass on session to share
        self.auth = oras.auth.get_auth_backend(auth_backend, self.session)

    @decorator.retry(attempts=5)
    def do_request(
        self,
        url: str,
        method: str = "GET",
        data: dict | bytes | None = None,
        headers: dict | None = None,
        json_param: dict | None = None,
        stream: bool = False,
    ) -> Response:
        """Do a request. This is a wrapper around requests to handle retry auth."""
        headers = headers or {}

        # Make the request and return to calling function, unless requires auth
        response = self.session.request(
            method,
            url,
            data=data,
            json=json_param,
            headers=headers,
            stream=stream,
            verify=self._tls_verify,
        )

        # A 401 response is a request for authentication
        if response.status_code not in [401, 404]:
            return response

        # Otherwise, authenticate the request and retry
        if self.authenticate_request(response):
            headers.update(self.headers)
            response = self.session.request(
                method,
                url,
                data=data,
                json=json_param,
                headers=headers,
                stream=stream,
            )

        return response

    def authenticate_request(self, original_response: Response) -> bool:
        """Authenticate a request with a given response."""
        auth_header_raw = original_response.headers.get("Www-Authenticate")
        if not auth_header_raw:
            self.logger.debug("Www-Authenticate not found in original response, cannot authenticate.")

            return False

        # If we have a token, set auth header (base64 encoded user/pass)
        if self.token:
            self.set_header("Authorization", f"Bearer {self.auth._basic_auth}")

            return True

        headers = copy.deepcopy(self.headers)
        auth_headers = oras.auth.utils.parse_auth_header(auth_header_raw)

        params = {}

        # Prepare request to retry
        if auth_headers.service:
            self.logger.debug(f"Service: {auth_headers.service}")
            params["service"] = auth_headers.service
            headers.update({
                "Service": auth_headers.service,
                "Accept": "application/json",
                "User-Agent": "oras-py",
            })

        # Ensure the realm starts with http
        if auth_headers.realm and not auth_headers.realm.startswith("http"):
            auth_headers.realm = f"{self.prefix}://{auth_headers.realm}"

        # If the www-authenticate included a scope, honor it!
        if auth_headers.scope:
            self.logger.debug(f"Scope: {auth_headers.scope}")
            params["scope"] = auth_headers.scope

        auth_response = self.session.get(auth_headers.realm, headers=headers, params=params)
        if auth_response.status_code != 200:
            self.logger.debug(f"Auth response was not successful: {auth_response.text}")

            return False

        # Request the token
        info = auth_response.json()
        token = info.get("token") or info.get("access_token")

        # Set the token to the original request and retry
        self.headers.update({"Authorization": f"Bearer {token}"})

        return True

    @ensure_container
    def push_container(
        self,
        container: Container,
        archives: list,
        logger: HopprLogger,
    ) -> Response:
        """Given a list of layer metadata (paths and corresponding mediaType) push."""
        # Prepare a new manifest
        manifest = oras.oci.NewManifest()
        self.session.cookies.set_policy(BlockAll())

        self.upload_blobs(container, archives, manifest, logger)

        # Prepare manifest and config (add your custom annotations here)
        manifest["annotations"] = {
            "org.opencontainers.image.created": str(datetime.now()),
            "ArtifactType": "Oras OCI Bundle",
            "Documentation": "https://oras.land/cli/",
        }
        config_file = "/tmp/oras-conf.json"
        config_blob = {
            "os": str(platform.platform()),
            "HopctlVersion": hoppr.__version__,
        }
        self.setup_default_config(config_blob, config_file)
        conf, config_file = oras.oci.ManifestConfig(
            path=config_file, media_type="application/vnd.oci.image.config.v1+json"
        )
        conf["annotations"] = config_blob

        # Config is just another layer blob!
        logger.info(f"Uploading config to {container.uri}")
        logger.info(f"Config {conf}")
        logger.flush()
        response = self.upload_blob(config_file, container, conf)
        self._check_200_response(response)
        Path(config_file).unlink(missing_ok=True)

        # Final upload of the manifest
        manifest["config"] = conf
        # Try updating the manifest
        response = self.upload_manifest(manifest=manifest, container=container)
        self._check_200_response(response=response)

        return response

    def put_upload(
        self,
        blob: str,
        container: oras.container.Container,
        layer: dict,
        refresh_headers: bool = True,
    ) -> Response:
        """Upload to a registry via put."""
        # Start an upload session
        headers = {"Content-Type": "application/octet-stream"}

        if not refresh_headers:
            headers |= self.headers

        upload_url = f"{self.prefix}://{container.upload_blob_url()}"
        request = self.do_request(upload_url, "POST", headers=headers)

        # Location should be in the header
        session_url = self._get_location(request, container)
        if not session_url:
            raise ValueError(f"Issue retrieving session url: {request.json()}")

        # PUT to upload blob url
        headers = {
            "Content-Length": str(layer["size"]),
            "Content-Type": "application/octet-stream",
        } | self.headers

        blob_url = oras.utils.append_url_params(session_url, {"digest": layer["digest"]})
        with Path(blob).open("rb") as fd:
            response = self.do_request(
                blob_url,
                method="PUT",
                data=fd.read(),
                headers=headers,
            )

        return response

    def upload_blobs(
        self,
        container: Container,
        archives: list,
        manifest: dict,
        logger: HopprLogger,
    ) -> dict:
        """Upload individual layers to OCI registry."""
        # Upload files as blobs
        for item in archives:
            blob = item.get("path")
            media_type = item.get("media_type")
            annots = item.get("annotations") or {}

            if not blob or not Path(blob).exists():
                logger.info(f"Path {blob} does not exist or is not defined.")

                continue

            # Artifact title is basename or user defined
            blob_name = item.get("title") or Path(blob).name

            # Create a new layer from the blob
            layer = oras.oci.NewLayer(blob, media_type, is_dir=False)
            logger.info(f"Preparing layer {layer}")

            # Update annotations with title we will need for extraction
            annots.update({oras.defaults.annotation_title: blob_name})
            layer["annotations"] = annots

            # update the manifest with the new layer
            manifest["layers"].append(layer)

            # Upload the blob layer
            logger.info(f"Uploading {blob} to {container.uri}")
            logger.flush()
            response = self.upload_blob(blob, container, layer)
            self._check_200_response(response)

        return manifest

    def upload_blob(
        self,
        blob: str,
        container: container_type,
        layer: dict,
        do_chunked: bool = False,
        refresh_headers: bool = True,
    ) -> Response:
        """Prepare and upload a blob."""
        # Reauthorize for each upload
        self.set_header("Authorization", self.auth.get_auth_header()["Authorization"])

        blob = os.path.abspath(blob)  # noqa: PTH100
        container = self.get_container(container)

        # Chunked for large, otherwise POST and PUT
        # This is currently disabled unless the user asks for it, as
        # it doesn't seem to work for all registries
        if not do_chunked:
            response = self.put_upload(blob, container, layer, refresh_headers=refresh_headers)
        else:
            response = self.chunked_upload(blob, container, layer)

        # If we have an empty layer digest and the registry didn't accept, just return dummy successful response
        if response.status_code not in [200, 201, 202] and layer["digest"] == oras.defaults.blank_hash:
            response = Response()
            response.status_code = 200

        return response

    def upload_manifest(self, manifest: dict, container: Container) -> Response:
        """Read a manifest file and upload it."""
        # Reauthorize manifest upload
        self.set_header("Authorization", self.auth.get_auth_header()["Authorization"])

        jsonschema.validate(manifest, schema=oras.schemas.manifest)

        headers = {
            "Content-Type": oras.defaults.default_manifest_media_type,
            "Content-Length": str(len(manifest)),
        }

        hostname, *_ = container.registry.split(":")
        put_url = f"{self.prefix}://{hostname}/v2/{container.api_prefix}/manifests/{container.tag}"

        return self.do_request(put_url, "PUT", headers=headers, json_param=manifest)

    def setup_default_config(self, config_blob: dict, config_file: str):
        """Setup default oras object config."""
        Path(config_file).write_text(json.dumps(obj=config_blob, indent=4))


class BlockAll(cookiejar.CookiePolicy):
    """Block all cookies."""

    def _block_all(self, *_, **__) -> bool:  # pragma: no cover
        return False

    netscape = True
    return_ok = set_ok = domain_return_ok = path_return_ok = _block_all
    rfc2965 = hide_cookie2 = False



"""Collector plugin for apt packages."""

from __future__ import annotations

import fnmatch
import os
import shutil

from configparser import ConfigParser
from os import PathLike
from pathlib import Path
from subprocess import CalledProcessError
from typing import TYPE_CHECKING
from urllib.parse import urlparse

import jc

from pydantic import SecretStr

import hoppr.utils

from hoppr import Hash, __version__, cdx
from hoppr.base_plugins.collector import BatchCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_process, hoppr_rerunner
from hoppr.exceptions import HopprPluginError
from hoppr.models.credentials import Credentials
from hoppr.models.types import PurlType
from hoppr.result import Result

if TYPE_CHECKING:
    from collections.abc import Mapping

    from packageurl import PackageURL

    from hoppr.models import HopprContext
    from hoppr.models.manifest import Repository
    from hoppr.models.sbom import Component


class CollectAptPlugin(
    BatchCollectorPlugin,
    required_commands=["apt", "apt-cache"],
    supported_purl_types=["deb"],
    products=["deb/*"],
):
    """Collector plugin for apt packages."""

    def __init__(
        self,
        context: HopprContext,
        config: dict | None = None,
        config_dir: str | PathLike[str] = Path.cwd() / ".hoppr-apt",
    ) -> None:
        super().__init__(
            context=context,
            config=config,
        )

        self.manifest_repos: list[str] = []

        self.manifest_repos = [f"{repo.url}" for repo in self.context.repositories[PurlType.DEB]]

        proxy_args = self._repo_proxy()
        self.config_dir = Path(config_dir)

        self.apt_paths_dict: Mapping[str, Mapping] = {
            "etc": {
                "apt": {"auth.conf": None, "preferences.d": {}, "sources.list": None, "sources.list.d": {}},
                "cache": {"apt": {"archives": {"lock": None, "partial": {}}}},
            },
            "var": {
                "lib": {
                    "apt": {"lists": {"lock": None, "partial": {}}},
                    "dpkg": {"lock-frontend": None, "status": None},
                }
            },
        }

        self.base_command = [
            self.required_commands[0],
            f"--option=Dir={self.config_dir}",
            f"--option=Dir::State::status={self.config_dir / 'var' / 'lib' / 'dpkg' / 'status'}",
            *proxy_args,
        ]

        self.debug_args = [
            "--option=Debug::pkgProblemResolver=1",
            "--option=Debug::Acquire::http=1",
            "--option=Debug::Acquire::https=1",
            "--option=Debug::Acquire::gpgv=1",
            "--option=Debug::pkgAcquire=1",
            "--option=Debug::pkgAcquire::Auth=1",
            "--option=Debug::pkgAcquire::Worker=1",
            "--option=Debug::pkgPackageManager=1",
            "--option=Debug::sourceList=1",
        ]

    def _artifact_string(self, purl: PackageURL) -> str:
        arch = purl.qualifiers.get("arch")
        return f"{purl.name}{f':{arch}' if arch else ''}{f'={purl.version}' if purl.version else ''}"

    def _download_component(self, purl: PackageURL) -> None:
        artifact = self._artifact_string(purl)

        try:
            # Download the component
            download_result = self.run_command([
                *self.base_command,
                *(self.debug_args if self.get_logger().is_verbose() else []),
                "download",
                artifact,
            ])

            download_result.check_returncode()
        except CalledProcessError as ex:
            raise ex

    def _get_component_download_info(self, purl: PackageURL) -> tuple[str, str, str]:
        artifact = self._artifact_string(purl=purl)

        try:
            # Try getting component download URL
            url_result = self.run_command([
                *self.base_command,
                "--print-uris",
                "download",
                artifact,
            ])

            url_result.check_returncode()

            if len(url_result.stdout.decode("utf-8")) == 0:
                raise HopprPluginError
        except (CalledProcessError, HopprPluginError) as ex:
            msg = f"Failed to get download URL for component: '{purl}'"
            self.get_logger().debug(msg=msg, indent_level=2)
            raise ex

        # Take the first URL and download filename if multiple are returned
        # Apt download output format:
        #   '<download URL>' <URL-encoded download filename> <download size> SHA512:<package hash>
        # Extract the output into a tuple of the form (<download URL>, <URL-encoded download filename>). For example:
        #   ("http://archive.ubuntu.com/ubuntu/pool/main/g/git/git_2.34.1-1ubuntu1.5_amd64.deb",
        #    "git_1%3a2.34.1-1ubuntu1.5_amd64.deb")
        result = url_result.stdout.decode("utf-8").split("\n")[0]
        found_url, download_filename, _, file_checksum = result.split(" ")
        *_, file_checksum = file_checksum.split(":")

        self.get_logger().debug(msg=f"Found URL: {found_url}", indent_level=3)
        self.get_logger().debug(msg=f"Download filename: {download_filename}", indent_level=3)

        return found_url.strip("'"), download_filename, file_checksum

    def _get_download_url_path(self, purl: PackageURL) -> str:
        """Get the path segment of the component download URL using `apt-cache show`."""
        artifact = self._artifact_string(purl)

        command = [*self.base_command, *(self.debug_args if self.get_logger().is_verbose() else []), "show", artifact]
        command[0] = "apt-cache"

        cmd_result = self.run_command(command)

        try:
            cmd_result.check_returncode()
        except CalledProcessError as ex:
            raise ex

        pkg_info = jc.parse(parser_mod_name="ini", data=cmd_result.stdout.decode(encoding="utf-8"))

        if not isinstance(pkg_info, dict) or not isinstance(pkg_info["Filename"], str):
            raise TypeError("Parsed output not in the expected format.")

        return pkg_info["Filename"]

    def _get_found_repo(self, found_url: str) -> str | None:
        """Identify the repository associated with the specified URL."""
        return next((repo for repo in self.manifest_repos if found_url.startswith(repo)), None)

    def _populate_apt_folder_structure(self, apt_path: Path, path_dict: Mapping[str, Mapping | None]) -> None:
        for key, value in path_dict.items():
            # None type indicates file to create
            if value is None:
                apt_file = apt_path / key
                apt_file.touch(exist_ok=True, mode=0o644)
            # dict type indicates directory to create
            elif isinstance(value, dict):
                apt_dir = apt_path / key
                apt_dir.mkdir(exist_ok=True, parents=True)

                if len(value.items()) > 0:
                    self._populate_apt_folder_structure(apt_path=apt_dir, path_dict=value)
            else:
                raise TypeError("Value is not expected type.")

    def _populate_auth_conf(self, repo_list: list[Repository], file: Path) -> None:
        """Populate Apt authentication config file."""
        creds: list[str] = []
        for repo in repo_list:
            repo_credentials = Credentials.find(f"{repo.url}")
            if repo_credentials is not None and isinstance(repo_credentials.password, SecretStr):
                creds.append(
                    f"machine {repo.url} "
                    f"login {repo_credentials.username} "
                    f"password {repo_credentials.password.get_secret_value()}\n"
                )

        # Set restrictive permissions on Apt authentication config file
        file.chmod(mode=0o600)

        with file.open(mode="w+", encoding="utf-8") as auth_conf:
            auth_conf.write("\n".join(creds))

    def _populate_sources_list(self, repo_list: list[str], file: Path) -> None:
        """Populate sources list."""
        # Read data from /etc/os-release file
        parser = ConfigParser()

        with (Path("/") / "etc" / "os-release").open(mode="r", encoding="utf-8") as os_release:
            parser.read_string(string=f"[os-release]\n{os_release.read()}")

        version_codename = parser["os-release"]["VERSION_CODENAME"]

        sources: list[str] = []
        for repo in repo_list:
            for component in ["main restricted", "universe", "multiverse"]:
                sources.extend((
                    f"deb {repo} {version_codename} {component}",
                    f"deb {repo} {version_codename}-updates {component}",
                    f"deb {repo} {version_codename}-security {component}",
                ))

        with file.open(mode="w+", encoding="utf-8") as sources_list:
            sources_list.write("\n".join(sources))

    def _repo_proxy(self) -> set[str]:
        proxy_args: list[str] = []

        for proto in ["http", "https"]:
            if proxy := os.getenv(f"{proto}_proxy"):
                proxy_args = [*proxy_args, f"--option=Acquire::{proto}::Proxy={proxy}"]

        no_proxy_urls = [item for item in os.getenv("no_proxy", "").split(",") if item != ""]

        for url in self.manifest_repos:
            parsed_url = urlparse(url)

            for pattern in no_proxy_urls:
                # Check if pattern is a substring or wildcard match of manifest repo URL
                if pattern in parsed_url.netloc or fnmatch.fnmatch(name=parsed_url.netloc, pat=pattern):
                    proxy_args = [
                        *proxy_args,
                        f"--option=Acquire::http::Proxy::{parsed_url.netloc}=DIRECT",
                        f"--option=Acquire::https::Proxy::{parsed_url.netloc}=DIRECT",
                    ]

        return set(proxy_args)

    def get_version(self) -> str:  # noqa: D102
        return __version__

    @hoppr_process
    def pre_stage_process(self) -> Result:  # noqa: D102
        self._populate_apt_folder_structure(apt_path=self.config_dir, path_dict=self.apt_paths_dict)

        self._populate_sources_list(
            repo_list=self.manifest_repos,
            file=self.config_dir / "etc" / "apt" / "sources.list",
        )

        self._populate_auth_conf(
            repo_list=self.context.repositories[PurlType.DEB],
            file=self.config_dir / "etc" / "apt" / "auth.conf",
        )

        if not self.context.strict_repos:
            system_apt_path = Path("/") / "etc" / "apt"
            plugin_apt_path = self.config_dir / "etc" / "apt"

            # Copy system Apt source lists into temporary directory
            shutil.copyfile(
                src=system_apt_path / "sources.list",
                dst=plugin_apt_path / "sources.list.d" / "system.list",
            )

            shutil.copytree(
                src=system_apt_path / "sources.list.d",
                dst=plugin_apt_path / "sources.list.d",
                dirs_exist_ok=True,
            )

        # Populate user Apt cache
        result = self.run_command([
            *self.base_command,
            *(self.debug_args if self.get_logger().is_verbose() else []),
            "--option=Dir::Etc::trusted=/etc/apt/trusted.gpg",
            "--option=Dir::Etc::trustedparts=/etc/apt/trusted.gpg.d",
            "update",
        ])

        if result.returncode != 0:
            return Result.fail("Failed to populate Apt cache.")

        return Result.success()

    def _update_hashes(self, comp: Component, package_path: Path):
        """Update the hashes of a component with hashes generated from its package file.

        Args:
            comp: Component object whose hashes will be updated.
            package_path: Path of the downloaded package file for the provided component.

        Raises:
            HopprPluginError: If an existing hash doesn't match.
        """
        generated_hashes = self._get_artifact_hashes(package_path)
        try:
            comp.update_hashes(generated_hashes)
        except ValueError as ex:
            raise HopprPluginError(f"Hash for {comp.name} does not match expected hash.") from ex

    def _validate_hashes(self, comp: Component, file_checksum: str):
        validation_hash = Hash(alg=cdx.HashAlg.SHA_512, content=file_checksum)
        if not comp.contains_hashes([validation_hash]):
            raise HopprPluginError(f"Hash validation failed for {comp.name}.")

    @hoppr_rerunner
    def collect(self, comp: Component) -> Result:
        """Copy a component to the local collection directory structure."""
        purl = hoppr.utils.get_package_url(comp.purl)

        try:
            found_url, download_filename, file_checksum = self._get_component_download_info(purl=purl)
        except (CalledProcessError, HopprPluginError) as ex:
            return Result.retry(message=str(ex))

        download_url_path = Path(self._get_download_url_path(purl=purl))
        subdir = download_url_path.parent

        if self.context.strict_repos:
            # Strip surrounding single quotes to compare against manifest repos
            repo = self._get_found_repo(found_url)

            # Return failure if found APT URL is not from a repo defined in the manifest
            if repo is None:
                return Result.fail(
                    f"Apt download URL does not match any repository in manifest. (Found URL: '{found_url}')"
                )

            result = self.check_purl_specified_url(purl, repo)
            if not result.is_success():
                return result
        else:
            # Default to found_url with the path component removed
            repo = found_url.removesuffix(str(download_url_path))

        target_dir = self.directory_for(purl.type, repo, subdir=str(subdir))
        package_out_path = target_dir / download_filename

        try:
            self._download_component(purl=purl)
        except CalledProcessError:
            msg = f"Failed to download Apt artifact {purl.name} version {purl.version}"
            return Result.retry(msg)

        self.get_logger().info(msg="Moving downloaded component:", indent_level=2)
        self.get_logger().info(msg=f"source: {download_filename}", indent_level=3)
        self.get_logger().info(msg=f"destination: {target_dir}", indent_level=3)

        shutil.move(src=download_filename, dst=target_dir)

        try:
            self._update_hashes(comp, target_dir / download_filename)
            self._validate_hashes(comp, file_checksum)
        except HopprPluginError as ex:
            return Result.fail(message=str(ex))

        self.set_collection_params(comp, repo, package_out_path)

        return Result.success(return_obj=comp)
